# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/Users/martinrusilowicz/work/apps/bio42/bio42_gui/forms/designer/frm_import_csv_designer.ui'
#
# Created by: PyQt5 UI code generator 5.9
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def __init__(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(880, 599)
        Dialog.setStyleSheet("/**\n"
"    PURPOSE:\n"
"        This is the default style-sheet used by all Intermake dialogues.\n"
"        It needs to be processed by css_processing.py before it can be used. \n"
"        It can be retrieved in processed form by the `intermake_gui.default_style_sheet()` function.\n"
"\n"
"    USAGE:    \n"
"        You can replace this stylesheet with your own.\n"
"        If you blank the contents of this stylesheet, the OSs default controls will be used.\n"
"        If you delete this stylesheet, the program will crash.\n"
"        \n"
"    EXTENSIONS:\n"
"        Normally not permitted in Qt, the following values are read through Intermake.\n"
"            * #DEFINE X Y                    - replaces all text `X` with `Y`\n"
"            * #WHEN X Y Z                    - only executes the following lines if the current\n"
"                                               section is any of `X` `Y` or `Z`.\n"
"                                               The section is specified when the user selects a\n"
"                                               stylesheet.\n"
"            * `QApplication.style`           - one of the Qt styles\n"
"            * `QApplication.small_icon_size` - the menu icon size, permitted only if `style` is set\n"
"            * `QMdiArea.background`          - colour of the Mdi area\n"
"\n"
"    DETAILS:\n"
"        Follow standard Qt stylesheet guidelines.\n"
"        \n"
"        The `:root` section defines constants that may be used elsewhere. These constants are\n"
"        substituted during the the processing stage and the `:root` section is removed.\n"
"        \n"
"        Intermake controls may have a string property named \"theme\" assigned to to certain widgets.\n"
"        This specifies that a unique appearance for the widget is intended:\n"
"        \n"
"        WIDGET        | THEME             | APPEARANCE (GUIDE)            | USAGE (GUIDE)\n"
"        --------------+-------------------+-------------------------------+-------------------------------\n"
"        QLabel        | heading           | border, big, bold             | section titles \n"
"        QLabel        | subheading        | border, big, bold             | section titles \n"
"        QTextEdit     | console           | monospaced, black background  | code, console output\n"
"        QPushButton   | completed         |                               |\n"
"        QPushButton   | cancel            | red                           | abort button\n"
"        QFrame        | header            | border                        | section titles\n"
"        QFrame        | contents          | white                         | with objects in\n"
"        QToolButton   | listbutton        | condensed                     | buttons in lists\n"
"        QLabel        | helpbox           | tooltip background            | help labels\n"
"        QLabel        | icon              | background suitable for image | label showing an icon\n"
"        QLabel        | warning           | yellow background, red text   | warning messages     \n"
"        QMdiArea      | empty             | darker                        | when MDI area has no windows\n"
"        ﻿QToolButton   | combo             | white                         | when button presents a selection menu\n"
"*/\n"
"\n"
"\n"
"\n"
"QMdiArea\n"
"{\n"
"    background : #E0E0E0;\n"
"}\n"
"\n"
"\n"
"QToolButton[style=\"listbutton\"]\n"
"{\n"
"    background   : #40C0FF;\n"
"    border-style : outset;\n"
"    border-width : 2px;\n"
"    border-color : transparent;\n"
"}\n"
"\n"
"QToolButton[style=\"listbutton\"]::hover\n"
"{\n"
"    background   : #B0D5E8;\n"
"    border-color : blue;\n"
"}\n"
"\n"
"QToolButton[style=\"listbutton\"]::pressed\n"
"{\n"
"    background   : #0040C0;\n"
"    border-style : inset;\n"
"}\n"
"\n"
"QLabel[style=\"icon\"]\n"
"{\n"
"    background    : #EEEEEE;\n"
"    border-radius : 8px;\n"
"}\n"
"\n"
"QFrame[style=\"title\"]\n"
"{\n"
"    margin-top     : 16px;\n"
"    margin-bottom  : 4px;\n"
"    margin-left    : 0px;\n"
"    margin-right   : 0px;\n"
"    border-radius  : 0px;\n"
"    border-bottom  : 2px solid silver;\n"
"    border-left    : none;\n"
"    border-right   : none;\n"
"    border-top     : none;\n"
"    padding-top    : 2px;\n"
"    padding-bottom : 2px;\n"
"    padding-left   : -4px;\n"
"    padding-right  : 0px;\n"
"    color          : black;\n"
"    font-size      : 18px;\n"
"}\n"
"\n"
"QLabel[style=\"title\"], QFrame[style=\"title\"]\n"
"{\n"
"    background    : #EEEEEE;\n"
"    border-radius : 4px;\n"
"    margin        : 2px;\n"
"    padding       : 2px;\n"
"    color         : black;\n"
"    font-size     : 18px;\n"
"}\n"
"\n"
"QLabel[style=\"title-embeded\"]\n"
"{\n"
"    background : #EEEEEE;\n"
"    color      : black;\n"
"    font-size  : 18px;\n"
"}\n"
"\n"
"\n"
"\n"
"\n"
"QLabel[style=\"helpbox\"]\n"
"{\n"
"    background    : transparent;\n"
"    color         : steelblue;\n"
"    padding       : 2px;\n"
"    border-radius : 4px;\n"
"}\n"
"\n"
"QLabel[style=\"subheading\"]\n"
"{\n"
"    font-weight: bold;\n"
"    font-style: italic;\n"
"}\n"
"\n"
"QLabel[style=\"heading\"], QPushButton[style=\"heading\"]\n"
"{\n"
"    font-weight: bold;\n"
"    border-bottom  : 1px solid #404040;\n"
"    border-left    : none;\n"
"    border-right   : none;\n"
"    border-top     : none;\n"
"    color: #404040;\n"
"}\n"
"\n"
"\n"
"QTextEdit[style=\"console\"]\n"
"{\n"
"    font-family: \"Consolas\", monospace;\n"
"    background : black;\n"
"    color      : white;\n"
"}\n"
"\n"
"QTextEdit[style=\"monospaced\"]\n"
"{\n"
"    font-family: \"Consolas\", monospace;\n"
"}\n"
"\n"
"QPushButton[style=\"completed\"]\n"
"{\n"
"    background    : #00C080;\n"
"    border-color  : #00C080; \n"
"}\n"
"\n"
"QPushButton[style=\"cancel\"]\n"
"{\n"
"    background    : #C00000;\n"
"    color         : white;\n"
"    padding       : 8px;\n"
"    border-color  : white;\n"
"    border-width  : 1px;\n"
"    border-radius : 8px;\n"
"}\n"
"\n"
"QMdiArea[style=\"empty\"]\n"
"{\n"
"    background : #E0E0E0;\n"
"}\n"
"\n"
"QLabel[style=\"warning\"]\n"
"{\n"
"    background       : #FFFFD0;\n"
"    padding          : 8px;\n"
"    border-radius    : 8px;\n"
"    image            : url(\":/intermake/warning.svg\");\n"
"    image-position   : left;\n"
"    qproperty-indent : 24;\n"
"}\n"
"\n"
"QLabel[style=\"warning_no_icon\"]\n"
"{\n"
"    background       : #FFFF80;\n"
"    color            : #800000;          \n"
"    padding          : 8px;\n"
"    border-radius    : 8px;\n"
"}\n"
"\n"
"QToolButton[style=\"dropdown\"]\n"
"{\n"
"    qproperty-toolButtonStyle : ToolButtonTextBesideIcon;\n"
"    qproperty-icon            : url(:/intermake/dropdown.svg);\n"
"}\n"
"\n"
"QToolButton[style=\"refresh\"]\n"
"{\n"
"    qproperty-toolButtonStyle : ToolButtonTextBesideIcon;\n"
"    qproperty-icon            : url(:/intermake/refresh.svg);\n"
"}\n"
"\n"
"\n"
"QToolButton[style=\"combo\"]\n"
"{\n"
"    background: white;\n"
"}\n"
"\n"
"QFrame[style=\"contents\"]\n"
"{\n"
"    background: white;\n"
"}\n"
"")
        self.gridLayout_2 = QtWidgets.QGridLayout(Dialog)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_4.addItem(spacerItem)
        self.BTN_BACK = QtWidgets.QPushButton(Dialog)
        self.BTN_BACK.setObjectName("BTN_BACK")
        self.horizontalLayout_4.addWidget(self.BTN_BACK)
        self.BTN_NEXT = QtWidgets.QPushButton(Dialog)
        self.BTN_NEXT.setObjectName("BTN_NEXT")
        self.horizontalLayout_4.addWidget(self.BTN_NEXT)
        self.BTN_OK = QtWidgets.QPushButton(Dialog)
        self.BTN_OK.setObjectName("BTN_OK")
        self.horizontalLayout_4.addWidget(self.BTN_OK)
        self.gridLayout_2.addLayout(self.horizontalLayout_4, 1, 0, 1, 1)
        self.STACK_MAIN = QtWidgets.QStackedWidget(Dialog)
        self.STACK_MAIN.setObjectName("STACK_MAIN")
        self.page_4 = QtWidgets.QWidget()
        self.page_4.setObjectName("page_4")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.page_4)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.label_19 = QtWidgets.QLabel(self.page_4)
        self.label_19.setObjectName("label_19")
        self.verticalLayout_3.addWidget(self.label_19)
        self.label_20 = QtWidgets.QLabel(self.page_4)
        self.label_20.setObjectName("label_20")
        self.verticalLayout_3.addWidget(self.label_20)
        self.CMB_DESTINATION = QtWidgets.QComboBox(self.page_4)
        self.CMB_DESTINATION.setObjectName("CMB_DESTINATION")
        self.verticalLayout_3.addWidget(self.CMB_DESTINATION)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem1)
        self.BTN_DESTINATION = QtWidgets.QPushButton(self.page_4)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/bio42/parcel.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.BTN_DESTINATION.setIcon(icon)
        self.BTN_DESTINATION.setObjectName("BTN_DESTINATION")
        self.horizontalLayout_3.addWidget(self.BTN_DESTINATION)
        self.BTN_REFRESH_DESTINATION = QtWidgets.QPushButton(self.page_4)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/bio42/refresh.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.BTN_REFRESH_DESTINATION.setIcon(icon1)
        self.BTN_REFRESH_DESTINATION.setObjectName("BTN_REFRESH_DESTINATION")
        self.horizontalLayout_3.addWidget(self.BTN_REFRESH_DESTINATION)
        self.verticalLayout_3.addLayout(self.horizontalLayout_3)
        spacerItem2 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_3.addItem(spacerItem2)
        self.STACK_MAIN.addWidget(self.page_4)
        self.page = QtWidgets.QWidget()
        self.page.setObjectName("page")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.page)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label_13 = QtWidgets.QLabel(self.page)
        self.label_13.setObjectName("label_13")
        self.verticalLayout.addWidget(self.label_13)
        self.label_16 = QtWidgets.QLabel(self.page)
        self.label_16.setObjectName("label_16")
        self.verticalLayout.addWidget(self.label_16)
        self.label = QtWidgets.QLabel(self.page)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.CMB_FILENAME = QtWidgets.QComboBox(self.page)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.CMB_FILENAME.sizePolicy().hasHeightForWidth())
        self.CMB_FILENAME.setSizePolicy(sizePolicy)
        self.CMB_FILENAME.setEditable(True)
        self.CMB_FILENAME.setObjectName("CMB_FILENAME")
        self.horizontalLayout.addWidget(self.CMB_FILENAME)
        self.BTN_FILENAME = QtWidgets.QPushButton(self.page)
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/bio42/browse.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.BTN_FILENAME.setIcon(icon2)
        self.BTN_FILENAME.setObjectName("BTN_FILENAME")
        self.horizontalLayout.addWidget(self.BTN_FILENAME)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.label_8 = QtWidgets.QLabel(self.page)
        self.label_8.setObjectName("label_8")
        self.verticalLayout.addWidget(self.label_8)
        self.gridFrame = QtWidgets.QFrame(self.page)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.gridFrame.sizePolicy().hasHeightForWidth())
        self.gridFrame.setSizePolicy(sizePolicy)
        self.gridFrame.setObjectName("gridFrame")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.gridFrame)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.CMB_DELIMITER = QtWidgets.QComboBox(self.gridFrame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.CMB_DELIMITER.sizePolicy().hasHeightForWidth())
        self.CMB_DELIMITER.setSizePolicy(sizePolicy)
        self.CMB_DELIMITER.setObjectName("CMB_DELIMITER")
        self.gridLayout_3.addWidget(self.CMB_DELIMITER, 0, 1, 1, 1)
        self.label_10 = QtWidgets.QLabel(self.gridFrame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_10.sizePolicy().hasHeightForWidth())
        self.label_10.setSizePolicy(sizePolicy)
        self.label_10.setObjectName("label_10")
        self.gridLayout_3.addWidget(self.label_10, 0, 0, 1, 1)
        self.CMB_ARRAY_DELIMITER = QtWidgets.QComboBox(self.gridFrame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.CMB_ARRAY_DELIMITER.sizePolicy().hasHeightForWidth())
        self.CMB_ARRAY_DELIMITER.setSizePolicy(sizePolicy)
        self.CMB_ARRAY_DELIMITER.setObjectName("CMB_ARRAY_DELIMITER")
        self.gridLayout_3.addWidget(self.CMB_ARRAY_DELIMITER, 1, 1, 1, 1)
        self.label_21 = QtWidgets.QLabel(self.gridFrame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_21.sizePolicy().hasHeightForWidth())
        self.label_21.setSizePolicy(sizePolicy)
        self.label_21.setObjectName("label_21")
        self.gridLayout_3.addWidget(self.label_21, 1, 0, 1, 1)
        self.verticalLayout.addWidget(self.gridFrame)
        self.CHK_F_HEADERS = QtWidgets.QCheckBox(self.page)
        self.CHK_F_HEADERS.setObjectName("CHK_F_HEADERS")
        self.verticalLayout.addWidget(self.CHK_F_HEADERS)
        self.CHK_F_TYPE_INFO = QtWidgets.QCheckBox(self.page)
        self.CHK_F_TYPE_INFO.setObjectName("CHK_F_TYPE_INFO")
        self.verticalLayout.addWidget(self.CHK_F_TYPE_INFO)
        self.CHK_F_FILENAME_LABELS = QtWidgets.QCheckBox(self.page)
        self.CHK_F_FILENAME_LABELS.setObjectName("CHK_F_FILENAME_LABELS")
        self.verticalLayout.addWidget(self.CHK_F_FILENAME_LABELS)
        spacerItem3 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem3)
        self.STACK_MAIN.addWidget(self.page)
        self.page_2 = QtWidgets.QWidget()
        self.page_2.setObjectName("page_2")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.page_2)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label_2 = QtWidgets.QLabel(self.page_2)
        self.label_2.setObjectName("label_2")
        self.verticalLayout_2.addWidget(self.label_2)
        self.label_17 = QtWidgets.QLabel(self.page_2)
        self.label_17.setObjectName("label_17")
        self.verticalLayout_2.addWidget(self.label_17)
        self.LST_ENTITIES = QtWidgets.QTreeWidget(self.page_2)
        self.LST_ENTITIES.setObjectName("LST_ENTITIES")
        self.LST_ENTITIES.headerItem().setText(0, "1")
        self.verticalLayout_2.addWidget(self.LST_ENTITIES)
        self.FRA_ADD_EDGES = QtWidgets.QFrame(self.page_2)
        self.FRA_ADD_EDGES.setObjectName("FRA_ADD_EDGES")
        self.gridLayout_5 = QtWidgets.QGridLayout(self.FRA_ADD_EDGES)
        self.gridLayout_5.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_5.setVerticalSpacing(0)
        self.gridLayout_5.setObjectName("gridLayout_5")
        self.LBL_ADD_ENTITY_2 = QtWidgets.QLabel(self.FRA_ADD_EDGES)
        self.LBL_ADD_ENTITY_2.setObjectName("LBL_ADD_ENTITY_2")
        self.gridLayout_5.addWidget(self.LBL_ADD_ENTITY_2, 0, 0, 1, 5)
        self.LBL_ENT_DEST = QtWidgets.QLabel(self.FRA_ADD_EDGES)
        self.LBL_ENT_DEST.setAlignment(QtCore.Qt.AlignCenter)
        self.LBL_ENT_DEST.setObjectName("LBL_ENT_DEST")
        self.gridLayout_5.addWidget(self.LBL_ENT_DEST, 2, 3, 1, 1)
        self.CMB_EDGE_SOURCE = QtWidgets.QComboBox(self.FRA_ADD_EDGES)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.CMB_EDGE_SOURCE.sizePolicy().hasHeightForWidth())
        self.CMB_EDGE_SOURCE.setSizePolicy(sizePolicy)
        self.CMB_EDGE_SOURCE.setEditable(True)
        self.CMB_EDGE_SOURCE.setObjectName("CMB_EDGE_SOURCE")
        self.gridLayout_5.addWidget(self.CMB_EDGE_SOURCE, 3, 1, 1, 1)
        self.CMB_EDGE_LABEL = QtWidgets.QComboBox(self.FRA_ADD_EDGES)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.CMB_EDGE_LABEL.sizePolicy().hasHeightForWidth())
        self.CMB_EDGE_LABEL.setSizePolicy(sizePolicy)
        self.CMB_EDGE_LABEL.setEditable(True)
        self.CMB_EDGE_LABEL.setObjectName("CMB_EDGE_LABEL")
        self.gridLayout_5.addWidget(self.CMB_EDGE_LABEL, 3, 2, 1, 1)
        self.CMB_EDGE_DEST = QtWidgets.QComboBox(self.FRA_ADD_EDGES)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.CMB_EDGE_DEST.sizePolicy().hasHeightForWidth())
        self.CMB_EDGE_DEST.setSizePolicy(sizePolicy)
        self.CMB_EDGE_DEST.setEditable(True)
        self.CMB_EDGE_DEST.setObjectName("CMB_EDGE_DEST")
        self.gridLayout_5.addWidget(self.CMB_EDGE_DEST, 3, 3, 1, 1)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.RAD_EDGE = QtWidgets.QRadioButton(self.FRA_ADD_EDGES)
        self.RAD_EDGE.setObjectName("RAD_EDGE")
        self.horizontalLayout_2.addWidget(self.RAD_EDGE)
        self.RAD_NODE = QtWidgets.QRadioButton(self.FRA_ADD_EDGES)
        self.RAD_NODE.setObjectName("RAD_NODE")
        self.horizontalLayout_2.addWidget(self.RAD_NODE)
        spacerItem4 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem4)
        self.BTN_ADD_EDGE = QtWidgets.QPushButton(self.FRA_ADD_EDGES)
        self.BTN_ADD_EDGE.setObjectName("BTN_ADD_EDGE")
        self.horizontalLayout_2.addWidget(self.BTN_ADD_EDGE)
        self.BTN_REMOVE_ENTITY = QtWidgets.QPushButton(self.FRA_ADD_EDGES)
        self.BTN_REMOVE_ENTITY.setObjectName("BTN_REMOVE_ENTITY")
        self.horizontalLayout_2.addWidget(self.BTN_REMOVE_ENTITY)
        self.gridLayout_5.addLayout(self.horizontalLayout_2, 1, 1, 1, 4)
        self.LBL_ENT_LABEL = QtWidgets.QLabel(self.FRA_ADD_EDGES)
        self.LBL_ENT_LABEL.setAlignment(QtCore.Qt.AlignCenter)
        self.LBL_ENT_LABEL.setObjectName("LBL_ENT_LABEL")
        self.gridLayout_5.addWidget(self.LBL_ENT_LABEL, 2, 2, 1, 1)
        self.LBL_ENT_SOURCE = QtWidgets.QLabel(self.FRA_ADD_EDGES)
        self.LBL_ENT_SOURCE.setAlignment(QtCore.Qt.AlignCenter)
        self.LBL_ENT_SOURCE.setObjectName("LBL_ENT_SOURCE")
        self.gridLayout_5.addWidget(self.LBL_ENT_SOURCE, 2, 1, 1, 1)
        self.verticalLayout_2.addWidget(self.FRA_ADD_EDGES)
        self.STACK_MAIN.addWidget(self.page_2)
        self.page_3 = QtWidgets.QWidget()
        self.page_3.setObjectName("page_3")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.page_3)
        self.verticalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.label_7 = QtWidgets.QLabel(self.page_3)
        self.label_7.setObjectName("label_7")
        self.verticalLayout_4.addWidget(self.label_7)
        self.label_18 = QtWidgets.QLabel(self.page_3)
        self.label_18.setObjectName("label_18")
        self.verticalLayout_4.addWidget(self.label_18)
        self.TVW_FIELDS = QtWidgets.QTreeWidget(self.page_3)
        self.TVW_FIELDS.setRootIsDecorated(False)
        self.TVW_FIELDS.setItemsExpandable(False)
        self.TVW_FIELDS.setObjectName("TVW_FIELDS")
        self.TVW_FIELDS.headerItem().setText(0, "1")
        self.verticalLayout_4.addWidget(self.TVW_FIELDS)
        self.label_15 = QtWidgets.QLabel(self.page_3)
        self.label_15.setObjectName("label_15")
        self.verticalLayout_4.addWidget(self.label_15)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.LBL_E_TYPE = QtWidgets.QLabel(self.page_3)
        self.LBL_E_TYPE.setObjectName("LBL_E_TYPE")
        self.gridLayout.addWidget(self.LBL_E_TYPE, 0, 1, 1, 1)
        self.CMB_E_TYPE = QtWidgets.QComboBox(self.page_3)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.CMB_E_TYPE.sizePolicy().hasHeightForWidth())
        self.CMB_E_TYPE.setSizePolicy(sizePolicy)
        self.CMB_E_TYPE.setObjectName("CMB_E_TYPE")
        self.gridLayout.addWidget(self.CMB_E_TYPE, 1, 1, 1, 1)
        self.CMB_ENTITY = QtWidgets.QComboBox(self.page_3)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.CMB_ENTITY.sizePolicy().hasHeightForWidth())
        self.CMB_ENTITY.setSizePolicy(sizePolicy)
        self.CMB_ENTITY.setObjectName("CMB_ENTITY")
        self.gridLayout.addWidget(self.CMB_ENTITY, 1, 0, 1, 1)
        self.LBL_E_NAME = QtWidgets.QLabel(self.page_3)
        self.LBL_E_NAME.setObjectName("LBL_E_NAME")
        self.gridLayout.addWidget(self.LBL_E_NAME, 0, 2, 1, 1)
        self.label_9 = QtWidgets.QLabel(self.page_3)
        self.label_9.setObjectName("label_9")
        self.gridLayout.addWidget(self.label_9, 0, 0, 1, 1)
        self.CMB_E_NAME = QtWidgets.QComboBox(self.page_3)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.CMB_E_NAME.sizePolicy().hasHeightForWidth())
        self.CMB_E_NAME.setSizePolicy(sizePolicy)
        self.CMB_E_NAME.setEditable(True)
        self.CMB_E_NAME.setObjectName("CMB_E_NAME")
        self.gridLayout.addWidget(self.CMB_E_NAME, 1, 2, 1, 1)
        self.verticalLayout_4.addLayout(self.gridLayout)
        self.STACK_MAIN.addWidget(self.page_3)
        self.page_5 = QtWidgets.QWidget()
        self.page_5.setObjectName("page_5")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.page_5)
        self.verticalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.label_11 = QtWidgets.QLabel(self.page_5)
        self.label_11.setObjectName("label_11")
        self.verticalLayout_5.addWidget(self.label_11)
        self.LBL_REVIEW = QtWidgets.QTextBrowser(self.page_5)
        self.LBL_REVIEW.setObjectName("LBL_REVIEW")
        self.verticalLayout_5.addWidget(self.LBL_REVIEW)
        self.STACK_MAIN.addWidget(self.page_5)
        self.gridLayout_2.addWidget(self.STACK_MAIN, 0, 0, 1, 1)

        self.retranslateUi(Dialog)
        self.STACK_MAIN.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Dialog"))
        self.BTN_BACK.setText(_translate("Dialog", "Back"))
        self.BTN_NEXT.setText(_translate("Dialog", "Next"))
        self.BTN_OK.setText(_translate("Dialog", "OK"))
        self.label_19.setText(_translate("Dialog", "Destination"))
        self.label_19.setProperty("style", _translate("Dialog", "title"))
        self.label_20.setText(_translate("Dialog", "Specify where to send your data to.\n"
"For large datasets you are recommended to send your data to a Parcel first."))
        self.label_20.setProperty("style", _translate("Dialog", "helpbox"))
        self.CMB_DESTINATION.setToolTip(_translate("Dialog", "Specify the name of your parcel.\n"
"Click the button to create a new parcel."))
        self.BTN_DESTINATION.setToolTip(_translate("Dialog", "Click to create a new parcel."))
        self.BTN_DESTINATION.setText(_translate("Dialog", "New parcel"))
        self.BTN_REFRESH_DESTINATION.setToolTip(_translate("Dialog", "Click to create a new parcel."))
        self.BTN_REFRESH_DESTINATION.setText(_translate("Dialog", "Refresh"))
        self.label_13.setText(_translate("Dialog", "File"))
        self.label_13.setProperty("style", _translate("Dialog", "title"))
        self.label_16.setText(_translate("Dialog", "Specify the file you wish to import. This can be any CSV or TSV file, with or without headers and with or without Neo4j type information."))
        self.label_16.setProperty("style", _translate("Dialog", "helpbox"))
        self.label.setText(_translate("Dialog", "Name"))
        self.CMB_FILENAME.setToolTip(_translate("Dialog", "Specify the file from which you wish to import data."))
        self.BTN_FILENAME.setToolTip(_translate("Dialog", "Click to browse your disk."))
        self.BTN_FILENAME.setText(_translate("Dialog", "Browse"))
        self.label_8.setText(_translate("Dialog", "Type"))
        self.CMB_DELIMITER.setToolTip(_translate("Dialog", "<html><body>\n"
"Specify the file delimiter. Typically <tt>,</tt> or <tt>TAB</tt>.\n"
"</body></html>"))
        self.label_10.setText(_translate("Dialog", "Delimiter"))
        self.CMB_ARRAY_DELIMITER.setToolTip(_translate("Dialog", "<html><body>\n"
"If your file contains arrays, specify the delimiter here.\n"
"Typically <tt>;</tt>.\n"
"This is only used if you specify fields of type array (<tt>[]</tt>).\n"
"</body></html>"))
        self.label_21.setText(_translate("Dialog", "Array delimiter"))
        self.CHK_F_HEADERS.setToolTip(_translate("Dialog", "Select this option if your file contains headers.\n"
"\n"
"File headers can be any arbitrary information that you wish to use to specify the columns.\n"
"If your file contains headers, you must select this option or the first row will be treated as data."))
        self.CHK_F_HEADERS.setText(_translate("Dialog", "File contains headers"))
        self.CHK_F_TYPE_INFO.setToolTip(_translate("Dialog", "<html><body>\n"
"<p>\n"
"Select this option if your file\'s headers contain type information.\n"
"</p>\n"
"<p>\n"
"If they don\'t, you will be prompted to specify the datatypes manually.\n"
"</p>\n"
"\n"
"<p>\n"
"Type information should be of the form: <tt>entity.property:type</tt>\n"
"<ul>\n"
"<li><tt>entity</tt> - the entity, specify either a label, such as <tt>Sequence</tt> or <tt>LIKE</tt>, or the fully decorated name, such as <tt>(Sequence)</tt> or <tt>(Sequence)-[LIKE]->(Sequence)</tt></li>\n"
"<li><tt>property</tt> - the name of the property. You can also use the special values: <tt>UID</tt> <tt>START_ID(label)</tt> and <tt>END_ID(label)</tt>.\n"
"<li><tt>type</tt> - the type of the property, such as <tt>int</tt> or <tt>str[]</tt>.\n"
"</ul>\n"
"</p>\n"
"</body></html>"))
        self.CHK_F_TYPE_INFO.setText(_translate("Dialog", "Headers contain type information"))
        self.CHK_F_FILENAME_LABELS.setToolTip(_translate("Dialog", "<html><body>\n"
"<p>\n"
"Select this option if your filename contains entity labels.\n"
"</p><p>\n"
"If it doesn\'t, you will be prompted to specify the entities manually.\n"
"</p>\n"
"<p>\n"
"Valid filenames are of the form <tt>Node-label.extension</tt> or <tt>Edge-label-label-label.extension</tt>.\n"
"</p>\n"
"<p>\n"
"For example <tt>Node-Sequence.csv</tt> or <tt>Edge-Sequence-LIKE-Sequence.csv</tt>.\n"
"</p>\n"
"</body></html>"))
        self.CHK_F_FILENAME_LABELS.setText(_translate("Dialog", "File name contains labels"))
        self.label_2.setText(_translate("Dialog", "Entities"))
        self.label_2.setProperty("style", _translate("Dialog", "title"))
        self.label_17.setText(_translate("Dialog", "Specify the labels of the nodes and edges that you will be importing.\n"
"If your filename is formatted appropriately this section will have already been filled out. Please review before continuing."))
        self.label_17.setProperty("style", _translate("Dialog", "helpbox"))
        self.LBL_ADD_ENTITY_2.setText(_translate("Dialog", "Configure"))
        self.LBL_ADD_ENTITY_2.setProperty("style", _translate("Dialog", "heading"))
        self.LBL_ENT_DEST.setText(_translate("Dialog", "Destination"))
        self.RAD_EDGE.setText(_translate("Dialog", "Edge"))
        self.RAD_NODE.setText(_translate("Dialog", "Node"))
        self.BTN_ADD_EDGE.setText(_translate("Dialog", "Duplicate"))
        self.BTN_REMOVE_ENTITY.setText(_translate("Dialog", "Remove"))
        self.LBL_ENT_LABEL.setText(_translate("Dialog", "Label"))
        self.LBL_ENT_SOURCE.setText(_translate("Dialog", "Source"))
        self.label_7.setText(_translate("Dialog", "Columns"))
        self.label_7.setProperty("style", _translate("Dialog", "title"))
        self.label_18.setText(_translate("Dialog", "Specify the fields of your CSV file.\n"
"If your CSV file already contains Neo4j type information this section will have already been filled out. Please review before continuing."))
        self.label_18.setProperty("style", _translate("Dialog", "helpbox"))
        self.label_15.setText(_translate("Dialog", "Configure"))
        self.label_15.setProperty("style", _translate("Dialog", "heading"))
        self.LBL_E_TYPE.setText(_translate("Dialog", "Type"))
        self.LBL_E_NAME.setText(_translate("Dialog", "Name"))
        self.label_9.setText(_translate("Dialog", "Entity"))
        self.label_11.setText(_translate("Dialog", "Review"))
        self.label_11.setProperty("style", _translate("Dialog", "title"))





from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QMessageBox, QWidget
from bio42_gui.forms.designer import frm_import_csv_designer
from bio42_gui.forms.frm_b42_base import AbstractB42SubWindow
from mhelper import string_helper, file_helper
from mhelper_qt import exceptToGui, exqtSlot, qt_gui_helper, tree_helper
from typing import List, Optional

import os.path
import csv
import bio42
import intermake
import neocommand as nc





class FrmImportCsv( AbstractB42SubWindow ):
    """
    Guides the user through importing a CSV using the `import_csv` command.
    """
    
    
    class Settings:
        """
        :ivar history: List of CSV files previously imported.
        """
        
        def __init__( self ):
            self.history: List[str] = []
    
    
    def __init__( self, parent: QWidget ):
        """
        CONSTRUCTOR
        """
        super().__init__( parent )
        self.ui = frm_import_csv_designer.Ui_Dialog( self )
        self.setWindowTitle( "Import CSV" )
        self.settings = intermake.Controller.ACTIVE.app.local_data.store.bind( type( self ).__name__, self.Settings() )
        
        # Delimiters
        for x in (self.ui.CMB_DELIMITER, self.ui.CMB_ARRAY_DELIMITER):
            x.addItem( "," )
            x.addItem( "TAB" )
            x.addItem( ";" )
            x.addItem( "|" )
        
        self.ui.CMB_DELIMITER.setCurrentText( "," )
        self.ui.CMB_ARRAY_DELIMITER.setCurrentText( ";" )
        self.ui.CHK_F_HEADERS.setChecked( True )
        
        # Inputs
        self.ui.CMB_FILENAME.addItems( self.settings.history )
        self.handle_update_filename()
        
        # Destinations
        self.__update_destinations()
        
        s: nc.DatabaseSchema = nc.get_core().schema
        
        # Properties
        self.ui.CMB_E_NAME.addItems( sorted( s.get_distinct_property_names() ) )
        
        # Types
        self.ui.CMB_E_TYPE.addItems( ["*{}".format( x.name ) for x in nc.ENcvSpecial if x is not nc.ENcvSpecial.NONE] )
        self.ui.CMB_E_TYPE.addItems( sorted( x.neo4j_name for x in nc.NeoType.ALLA ) )
        
        # Fields
        self.ui.CMB_EDGE_DEST.addItems( sorted( x.label for x in s.node_schema ) )
        self.ui.CMB_EDGE_SOURCE.addItems( sorted( x.label for x in s.node_schema ) )
        self.ui.CMB_EDGE_LABEL.addItems( sorted( set( x.label for x in s.iter_entity_types() ) ) )
        
        # Fields
        self.prev_index = -1
        self.ignore_field_changes = False
        self.ignore_entity_changes = False
        
        # Display
        self.ui.STACK_MAIN.setCurrentIndex( 0 )
        
        # Signals
        # Stack widget 
        self.ui.STACK_MAIN.currentChanged[int].connect( self._on_current_changed )
        
        # Entity editor
        self.ui.LST_ENTITIES.itemSelectionChanged.connect( self.entity_to_ui )
        self.ui.CMB_EDGE_DEST.currentTextChanged.connect( self.entity_from_ui )
        self.ui.CMB_EDGE_LABEL.currentTextChanged.connect( self.entity_from_ui )
        self.ui.CMB_EDGE_SOURCE.currentTextChanged.connect( self.entity_from_ui )
        self.ui.RAD_EDGE.toggled[bool].connect( self.entity_from_ui )
        self.ui.RAD_NODE.toggled[bool].connect( self.entity_from_ui )
        
        # Field editor
        self.ui.TVW_FIELDS.itemSelectionChanged.connect( self.field_to_ui )
        self.ui.CMB_ENTITY.currentTextChanged.connect( self.field_from_ui )
        self.ui.CMB_E_NAME.currentTextChanged.connect( self.field_from_ui )
        self.ui.CMB_E_TYPE.currentTextChanged.connect( self.field_from_ui )
    
    
    def on_plugin_completed( self, result: intermake.Result ):
        self.__update_destinations()
    
    
    def __update_destinations( self ):
        self.ui.CMB_DESTINATION.clear()
        self.ui.CMB_DESTINATION.addItems( [x.name for x in nc.get_core().endpoint_manager] )
    
    
    def entity_to_ui( self ):
        if self.ignore_entity_changes:
            return
        
        ent: _Entity = self.get_selected_entity()
        
        if ent is None:
            return
        
        self.ignore_entity_changes = True
        self.ui.RAD_EDGE.setChecked( ent.is_edge )
        self.ui.RAD_NODE.setChecked( not ent.is_edge )
        self.ui.CMB_EDGE_SOURCE.setCurrentText( ent.start_label )
        self.ui.CMB_EDGE_DEST.setCurrentText( ent.end_label )
        self.ui.CMB_EDGE_LABEL.setCurrentText( ent.label )
        
        self.ignore_entity_changes = False
        self.entity_update_buttons()
    
    
    def entity_from_ui( self, *_, **__ ):
        ent: _Entity = self.get_selected_entity()
        
        if ent is None:
            return
        
        ent.is_edge = self.ui.RAD_EDGE.isChecked()
        ent.start_label = self.ui.CMB_EDGE_SOURCE.currentText()
        ent.label = self.ui.CMB_EDGE_LABEL.currentText()
        ent.end_label = self.ui.CMB_EDGE_DEST.currentText()
        
        self.ent_to_list( ent )
        self.entity_update_buttons()
    
    
    def entity_update_buttons( self ):
        ent = self.get_selected_entity()
        
        if ent is None:
            return
        
        self.ui.LBL_ENT_DEST.setEnabled( ent.is_edge )
        self.ui.CMB_EDGE_DEST.setEnabled( ent.is_edge )
        self.ui.LBL_ENT_SOURCE.setEnabled( ent.is_edge )
        self.ui.CMB_EDGE_SOURCE.setEnabled( ent.is_edge )
    
    
    def field_from_ui( self ):
        if self.ignore_field_changes:
            return
        
        field = self.get_selected_field()
        
        if field is None:
            return
        
        entity_str = self.ui.CMB_ENTITY.currentText()
        
        if entity_str == "*NONE":
            field.entity = None
        else:
            for entity in tree_helper.iter_data( self.ui.LST_ENTITIES ):
                if str( entity ) == entity_str:
                    field.entity = entity
                    break
        
        name = self.ui.CMB_E_NAME.currentText()
        type_str = self.ui.CMB_E_TYPE.currentText()
        
        if type_str.startswith( "*" ):
            try:
                special = nc.ENcvSpecial[type_str[1:]]
            except KeyError:
                special = nc.ENcvSpecial.UID
            
            field.field = nc.NcvHeader( None, None, special, None )
        else:
            try:
                type_ = nc.NeoType.from_name( type_str )
            except KeyError:
                type_ = nc.NeoType.STR
            
            field.field = nc.NcvHeader( name, type_, nc.ENcvSpecial.NONE, None )
        
        self.field_to_list( field )
        self.field_update_buttons()
    
    
    @exceptToGui()
    def field_to_ui( self ):
        field = self.get_selected_field()
        
        if field is None:
            return
        
        self.ignore_field_changes = True
        
        self.ui.label_15.setText( "Configure: {}".format( field.column ) )
        self.ui.CMB_ENTITY.setCurrentText( str( field.entity ) if field.entity is not None else "*NONE" )
        self.ui.CMB_E_NAME.setCurrentText( field.field.name )
        
        if field.field.special == nc.ENcvSpecial.NONE:
            self.ui.CMB_E_TYPE.setCurrentText( field.field.type.neo4j_name )
        else:
            self.ui.CMB_E_TYPE.setCurrentText( "*{}".format( field.field.special.name ) )
        
        self.ignore_field_changes = False
        self.field_update_buttons()
    
    
    def field_update_buttons( self ):
        field = self.get_selected_field()
        
        if field is None:
            return
        
        self.ui.CMB_E_TYPE.setEnabled( field.entity is not None )
        self.ui.LBL_E_TYPE.setEnabled( field.entity is not None )
        self.ui.CMB_E_NAME.setEnabled( field.entity is not None and field.field.special == nc.ENcvSpecial.NONE )
        self.ui.LBL_E_NAME.setEnabled( field.entity is not None and field.field.special == nc.ENcvSpecial.NONE )
    
    
    def field_to_list( self, field: "_Field" ):
        item = tree_helper.get_or_create_item( self.ui.TVW_FIELDS, field )
        item.setText( tree_helper.get_or_create_column( self.ui.TVW_FIELDS, "Index" ), str( field.index ) )
        item.setText( tree_helper.get_or_create_column( self.ui.TVW_FIELDS, "Header" ), str( field.column ) )
        item.setText( tree_helper.get_or_create_column( self.ui.TVW_FIELDS, "Example" ), field.sample )
        
        if field.entity is not None:
            item.setText( tree_helper.get_or_create_column( self.ui.TVW_FIELDS, "Entity" ), str( field.entity ) )
            
            if field.field.special != nc.ENcvSpecial.NONE:
                item.setText( tree_helper.get_or_create_column( self.ui.TVW_FIELDS, "Type" ), "*{}".format( field.field.special ) )
                item.setText( tree_helper.get_or_create_column( self.ui.TVW_FIELDS, "Name" ), "--" )
            else:
                item.setText( tree_helper.get_or_create_column( self.ui.TVW_FIELDS, "Type" ), str( field.field.type.neo4j_name ) )
                item.setText( tree_helper.get_or_create_column( self.ui.TVW_FIELDS, "Name" ), str( field.field.name ) )
        else:
            item.setText( tree_helper.get_or_create_column( self.ui.TVW_FIELDS, "Entity" ), "--" )
            item.setText( tree_helper.get_or_create_column( self.ui.TVW_FIELDS, "Type" ), "--" )
            item.setText( tree_helper.get_or_create_column( self.ui.TVW_FIELDS, "Name" ), "--" )
    
    
    def ent_to_list( self, ent: "_Entity" ):
        item = tree_helper.get_or_create_item( self.ui.LST_ENTITIES, ent )
        item.setText( tree_helper.get_or_create_column( self.ui.LST_ENTITIES, "Type" ), "Edge" if ent.is_edge else "Node" )
        item.setText( tree_helper.get_or_create_column( self.ui.LST_ENTITIES, "Start" ), ent.start_label if ent.is_edge else "" )
        item.setText( tree_helper.get_or_create_column( self.ui.LST_ENTITIES, "Label" ), ent.label )
        item.setText( tree_helper.get_or_create_column( self.ui.LST_ENTITIES, "End" ), ent.end_label if ent.is_edge else "" )
        
        item.setForeground( 0, QColor( 0, 0, 0 ) if ent.is_valid else QColor( 255, 0, 0 ) )
    
    
    def get_selected_field( self ) -> Optional["_Field"]:
        return tree_helper.get_selected_data( self.ui.TVW_FIELDS )
    
    
    def get_selected_entity( self ) -> Optional["_Entity"]:
        return tree_helper.get_selected_data( self.ui.LST_ENTITIES )
    
    
    @exceptToGui()
    def _on_current_changed( self, index: int ):
        is_advance = self.prev_index < index
        self.prev_index = index
        self.ui.BTN_BACK.setEnabled( self.ui.STACK_MAIN.currentIndex() != 0 )
        self.ui.BTN_NEXT.setEnabled( self.ui.STACK_MAIN.currentIndex() != self.ui.STACK_MAIN.count() - 1 )
        self.ui.BTN_OK.setEnabled( self.ui.STACK_MAIN.currentIndex() == self.ui.STACK_MAIN.count() - 1 )
        
        if is_advance:
            if index == 2:
                # Entities
                
                if not os.path.isfile( self.ui.CMB_FILENAME.currentText() ):
                    self.ui.STACK_MAIN.setCurrentIndex( 1 )
                    QMessageBox.warning( self, "Oops", "Select a valid file before continuing." )
                    return
                
                # Save filename 
                fn = self.ui.CMB_FILENAME.currentText()
                if os.path.isfile( fn ):
                    if fn in self.settings.history:
                        self.settings.history.remove( fn )
                    
                    self.settings.history.insert( 0, fn )
                    intermake.Controller.ACTIVE.app.local_data.store.commit( self.settings )
                
                # Populate defaults
                self.ui.LST_ENTITIES.clear()
                
                if self.ui.CHK_F_FILENAME_LABELS.isChecked():
                    nf = nc.NcvFilename.construct_from_file( self.ui.CMB_FILENAME.currentText() )
                    
                    if nf.is_edge:
                        self.ent_to_list( _Entity( False, nf.start_label, "", "" ) )
                        self.ent_to_list( _Entity( False, nf.end_label, "", "" ) )
                        self.ent_to_list( _Entity( True, nf.label, nf.start_label, nf.end_label ) )
                    else:
                        self.ent_to_list( _Entity( False, nf.label, "", "" ) )
                
                if self.ui.LST_ENTITIES.topLevelItemCount() == 0:
                    self.ent_to_list( _Entity( False, "", "", "" ) )
                
                # Select first
                self.ui.LST_ENTITIES.setCurrentItem( self.ui.LST_ENTITIES.topLevelItem( 0 ) )
            elif index == 3:
                # Fields
                # Entity list
                self.ui.CMB_ENTITY.clear()
                self.ui.CMB_ENTITY.addItem( "*NONE" )
                for entity in tree_helper.iter_data( self.ui.LST_ENTITIES ):
                    self.ui.CMB_ENTITY.addItem( str( entity ) )
                
                # Default fields
                self.ui.TVW_FIELDS.clear()
                
                with open( self.ui.CMB_FILENAME.currentText(), "r" ) as file:
                    headers = None
                    
                    for row in csv.reader( file, delimiter = self.get_delimiter() ):
                        if headers is None:
                            headers = row
                        else:
                            row0 = row
                            break
                    else:
                        headers = ()
                        row0 = ()
                
                if self.ui.CHK_F_HEADERS.isChecked():
                    for index, (header_, row0_) in enumerate( zip( headers, row0 ) ):
                        header_: str = header_.strip()
                        row0_: str = row0_.strip()
                        
                        if "." in header_:
                            entity_str, header = header_.split( ".", 1 )
                        else:
                            header = header_
                            entity_str = None
                        
                        for e in tree_helper.iter_data( self.ui.LST_ENTITIES ):
                            assert isinstance( e, _Entity )
                            
                            if str( e ) == entity_str:
                                entity = e
                                break
                            elif e.label == entity_str:
                                entity = e
                                break
                        
                        if self.ui.CHK_F_TYPE_INFO.isChecked():
                            hd = nc.NcvHeader.from_decorated_name( None, header )
                            self.field_to_list( _Field( index, header_, entity, nc.NcvHeader( hd.name, hd.type, hd.special ), row0_ ) )
                        else:
                            hd = header.lower().replace( " ", "_" )
                            self.field_to_list( _Field( index, header_, entity, nc.NcvHeader( hd, nc.NeoType.STR, nc.ENcvSpecial.NONE ), row0_ ) )
                else:
                    for index, header_ in enumerate( headers ):
                        header_: str = header_.strip()
                        self.field_to_list( _Field( index, "#{}".format( index ), None, nc.NcvHeader( "untitled_{}".format( index ), nc.NeoType.STR, nc.ENcvSpecial.NONE ), header_ ) )
                
                # Select first
                self.ui.TVW_FIELDS.setCurrentItem( self.ui.TVW_FIELDS.topLevelItem( 0 ) )
            elif index == 4:
                # Review
                r = []
                r.append( "<html><body>" )
                r.append( "<h2>Summary</h2>" )
                r.append( "<table border=1>" )
                r.append( "<tr><td><b>{}</b></td><td><b>{}</b></td><td><b>{}</b></td></tr>".format( "Column", "Entity", "Field" ) )
                invalid = []
                
                for field in tree_helper.iter_data( self.ui.TVW_FIELDS ):
                    assert isinstance( field, _Field )
                    try:
                        r.append( "<tr><td>{}. \"<i>{}</i>\"</td><td>{}</td><td>{}</td></tr>".format( field.index, field.column, field.entity, field.field.decorated_name( field.entity ) if field.entity else "-" ) )
                        field.validate()
                    except Exception as ex:
                        invalid.append( (field, repr( ex )) )
                
                r.append( "</table>" )
                
                for entity in tree_helper.iter_data( self.ui.LST_ENTITIES ):
                    assert isinstance( entity, _Entity )
                    try:
                        entity.validate()
                    except Exception as ex:
                        invalid.append( (entity, repr( ex )) )
                
                r.append( "<h2>Issues</h2>" )
                if invalid:
                    r.append( "<p>Cannot continue: At least one entity or field is not configured: {}</p>".format( string_helper.format_array( invalid, format = "“{}”".format ) ) )
                    self.ui.BTN_OK.setEnabled( False )
                else:
                    r.append( "<p>No issues.</p>" )
                
                r.append( "<h2>Fields data</h2>" )
                r.append( "<ul>" )
                r.append( "".join( "<li>“{}”</li>".format( x ) for x in self.get_field_text() ) )
                r.append( "</ul>" )
                r.append( "</body></html>" )
                
                self.ui.LBL_REVIEW.setText( "\n".join( r ) )
    
    
    def get_delimiter( self ):
        return self.process_delimiter( self.ui.CMB_DELIMITER.currentText() )
    
    
    def process_delimiter( self, d ):
        if d in ("TAB", "\\t"):
            d = "\t"
        elif d in ("SPACE", "\\s"):
            d = " "
        elif d in ("CR", "\\r"):
            d = "\r"
        
        if len( d ) != 1:
            raise ValueError( "«{}» is not an acceptable delimiter.".format( d ) )
        
        return d
    
    
    def get_array_delimiter( self ):
        return self.process_delimiter( self.ui.CMB_ARRAY_DELIMITER.currentText() )
    
    
    @exqtSlot()
    def on_BTN_BACK_clicked( self ) -> None:
        """
        Signal handler:
        """
        if QMessageBox.question( self,
                                 self.windowTitle(),
                                 "If you go back you will lose any changes on the current page?",
                                 QMessageBox.Ok | QMessageBox.Cancel
                                 ) != QMessageBox.Ok:
            return
        
        if self.ui.STACK_MAIN.currentIndex() > 0:
            self.ui.STACK_MAIN.setCurrentIndex( self.ui.STACK_MAIN.currentIndex() - 1 )
    
    
    @exqtSlot()
    def on_BTN_NEXT_clicked( self ) -> None:
        """
        Signal handler:
        """
        if self.ui.STACK_MAIN.currentIndex() < self.ui.STACK_MAIN.count() - 1:
            self.ui.STACK_MAIN.setCurrentIndex( self.ui.STACK_MAIN.currentIndex() + 1 )
    
    
    @exqtSlot()
    def on_BTN_OK_clicked( self ) -> None:
        """
        Signal handler:
        """
        file = self.ui.CMB_FILENAME.currentText()
        endpoint = nc.get_core().endpoint_manager.find_endpoint( self.ui.CMB_DESTINATION.currentText() )
        delimiter = self.get_delimiter()
        array_delimiter = self.get_array_delimiter()
        fields = self.get_field_text()
        
        intermake.acquire( bio42.import_csv, window = self ) \
            .run( file = file,
                  endpoint = endpoint,
                  fields = fields,
                  delimiter = delimiter,
                  array_delimiter = array_delimiter )
    
    
    def get_field_text( self ):
        fields = []
        
        for field in tree_helper.iter_data( self.ui.TVW_FIELDS ):
            assert isinstance( field, _Field )
            assert field.index == len( fields )
            fields.append( field.to_import_csv() )
        
        return fields
    
    
    @exqtSlot()
    def on_BTN_DESTINATION_clicked( self ) -> None:
        """
        Signal handler:
        """
        self.show_command( nc.open_parcel )
    
    
    @exqtSlot()
    def on_BTN_FILENAME_clicked( self ) -> None:
        """
        Signal handler:
        """
        if qt_gui_helper.browse_open_on_textbox( self.ui.CMB_FILENAME ):
            self.handle_update_filename()
    
    
    def handle_update_filename( self ):
        cfn = self.ui.CMB_FILENAME.currentText()
        
        if file_helper.get_extension( cfn ) == ".tsv":
            self.ui.CMB_DELIMITER.setCurrentText( "TAB" )
        elif file_helper.get_extension( cfn ) == ".csv":
            self.ui.CMB_DELIMITER.setCurrentText( "," )
    
    
    @exqtSlot()
    def on_BTN_REMOVE_ENTITY_clicked( self ) -> None:
        """
        Signal handler:
        """
        if self.ui.LST_ENTITIES.topLevelItemCount() == 1:
            return
        
        tree_helper.remove_selected( self.ui.LST_ENTITIES )
    
    
    @exqtSlot()
    def on_BTN_REFRESH_DESTINATION_clicked( self ) -> None:
        """
        Signal handler:
        """
        self.__update_destinations()
    
    
    @exqtSlot()
    def on_BTN_ADD_EDGE_clicked( self ) -> None:
        """
        Signal handler:
        """
        ent = self.get_selected_entity()
        
        if ent is None:
            return
        
        self.ent_to_list( _Entity( ent.is_edge, ent.label, ent.start_label, ent.end_label ) )


class _Entity( nc.NcvEntitySpec ):
    """
    Entity definition on the import CSV form.
    """
    
    
    @property
    def is_valid( self ):
        try:
            self.validate()
        except:
            return False
        else:
            return True
    
    
    def validate( self ):
        if not self.label:
            raise ValueError( "Missing label" )
        
        if self.is_edge:
            if not self.start_label:
                raise ValueError( "Missing start label" )
            
            if not self.end_label:
                raise ValueError( "Missing end label" )


class _Field:
    """
    Field definition on the import CSV form.
    """
    
    
    def __init__( self, index, column, entity: Optional[_Entity], field: nc.NcvHeader, sample: str ):
        self.index = index
        self.column = column
        self.entity = entity
        self.field = field
        self.sample = sample
    
    
    def validate( self ):
        if not self.entity:
            return
        
        if self.field.special != nc.ENcvSpecial.NONE:
            return
        
        if not self.field.name:
            raise ValueError( "Missing field name" )
        
        if not self.field.type:
            raise ValueError( "Missing field type" )
    
    
    def to_import_csv( self ):
        if self.entity is None:
            return ""
        else:
            return "{}.{}".format( self.entity.label, self.field.decorated_name( self.entity ) )


from typing import Iterable
import editorium
import neocommand


class DatabaseDriverEditor( editorium.AbstractEnumEditor ):
    def on_get_options( self, info: editorium.EditorInfo ) -> Iterable[object]:
        return neocommand.DbManager.DRIVER_CLASSES.keys()
    
    
    @classmethod
    def on_can_handle( cls, info: editorium.EditorInfo ) -> bool:
        return info.annotation.is_mannotation_of( neocommand.isDriverName )


def register_with( x: editorium.Editorium ):
    x.register( DatabaseDriverEditor )



"""
Contains the BLAST to $(APP_DISPLAY_NAME) importer/parser.
"""
from typing import Dict, List, Optional, Tuple
from uuid import uuid4

from bio42.application import app
from bio42 import schema

import intermake
from mhelper import EFileMode, isFilename, isOptional
import neocommand


__author__ = "Martin Rusilowicz"

_SPC_NONE_ = -1
_SPC_QUERY_UID = 0
_SPC_SUBJECT_UID = 1
_SPC_QUERY_START = 2
_SPC_QUERY_END = 3
_SPC_SUBJECT_START = 4
_SPC_SUBJECT_END = 5
_SPC_COUNT_ = 6

_EXT_DICTIONARY = ".dict"
_EXT_BLAST = ".blast"
_EXT_TSV = ".tsv"


@app.command(folder="import")
def import_blast( file_name: isFilename[EFileMode.READ, _EXT_BLAST, _EXT_TSV],
                  destination: neocommand.Destination,
                  column_names: Optional[List[str]] = None,
                  dictionary_file_name: isOptional[isFilename[EFileMode.READ, _EXT_DICTIONARY, _EXT_TSV]] = None,
                  dictionary_uid_column: int = -1,
                  deep_query: bool = False
                  ) -> None:
    """
    Imports BLAST data as edges and/or nodes in the graph.
    
    :param destination: Where to send the parsed data to.
    :param file_name: File to read from.
    :param column_names: Mapping of the columns in the TSV file. See the BLAST+ documentation, or see the `BLAST Columns` help topic for quick reference.
    :param dictionary_file_name: Dictionary of names (TSV file giving Original_name and BLAST_name for each line).
    :param dictionary_uid_column: When using the dictionary, if not '-1' this specifies that the original names are '|' delimited FASTA accessions and this column should be taken as the UID.
    :param deep_query: When `True`, additional nodes and edges will be created: Subsequence, Query
    """
    with destination.open_writer() as writer:
        if dictionary_uid_column != -1:
            print( "Taking the nth element where n = {} as the UID column.".format( dictionary_uid_column ) )
        
        if deep_query:
            print( "Deep query is {},".format( "ON" if deep_query else "OFF" ) )
        
        if dictionary_file_name:
            uid_dictionary = { }
            
            with intermake.pr.pr_action( "Iterating dictionary file" ) as action:
                with open( dictionary_file_name, "r" ) as dictionary_file:
                    for line in dictionary_file:
                        action.increment()
                        if line:
                            elements = line.split( "\t" )
                            
                            original_name = elements[0].strip()
                            blast_name = elements[1].strip()
                            
                            if dictionary_uid_column != -1:
                                original_name = original_name.split( "|" )[dictionary_uid_column]
                            
                            uid_dictionary[blast_name] = original_name
        else:
            uid_dictionary = None
        
        if not column_names:
            column_names = ["std"]
        
        if len( column_names ) == 1:
            column_names = BLASTAUTOMAP[column_names[0]][1]
        
        column_mapping = []
        db_names = set()
        
        for i, column_name in enumerate( column_names ):
            mapping = BLASTMAP[column_name]
            db_name = mapping[0]
            
            if db_name in db_names:
                raise ValueError( "Column mapping specifies «{}» twice.".format( db_name ) )
            
            db_names.add( db_name )
            
            column_mapping.append( mapping )
        
        print( "Column names are:\n* {}".format( "\n* ".join( x[1] for x in column_mapping ) ) )
        
        with intermake.pr.pr_action( "Iterating BLAST file" ) as action:
            with open( file_name, "r" ) as file:
                for line in file:
                    action.increment()
                    __yield_line( writer, line, file_name, column_mapping, uid_dictionary, deep_query )


# noinspection PyUnusedLocal
def __yield_line( endpoint: neocommand.Writer, tsv_line: str, file_name: str, column_mapping: List[Tuple[str, str, type]], uid_dictionary: Dict[str, str], deep_query: bool ) -> None:
    e = tsv_line.split( "\t" )
    
    properties = { }
    specials = [None] * _SPC_COUNT_  # type: List[str]
    
    for i, element in enumerate( e ):
        mapping = column_mapping[i]
        
        name = mapping[1]
        
        try:
            value = mapping[2]( element )
        except Exception as ex:
            raise ValueError( "Failed to convert «{}» («{}») to «{}»".format( element, name, mapping[2].__name__ ) ) from ex
        
        special = mapping[3]
        
        properties[name] = value
        
        if special != _SPC_NONE_:
            specials[special] = value
    
    if any( x is None for x in specials ):
        raise ValueError( "Missing a mandatory column." )
    
    query_uid = specials[_SPC_QUERY_UID]  # TODO: Used to have truncate_id here, still need
    subject_uid = specials[_SPC_SUBJECT_UID]
    
    if uid_dictionary is not None:
        query_uid = uid_dictionary[query_uid]
        subject_uid = uid_dictionary[subject_uid]
    
    # Fields: query acc., subject acc., % identity, alignment length, mismatches, gap opens, q. start, q. end, s. start, s. end, e-value, bit score
    query_start = specials[_SPC_QUERY_START]
    query_end = specials[_SPC_QUERY_END]
    subject_start = specials[_SPC_SUBJECT_START]
    subject_end = specials[_SPC_SUBJECT_END]
    
    endpoint.create_edge( label = schema.SequenceBlastqueriedSequence,
                          start_uid = query_uid,
                          end_uid = subject_uid,
                          properties = dict( properties ) )
    
    if deep_query:
        # Expand BLAST edge to BLAST node
        blast_uid = str( uuid4() )
        endpoint.create_node( label = schema.Blast,
                              uid = blast_uid,
                              properties = dict( properties ) )
        
        endpoint.create_edge( label = schema.BlastQueriesSequence,
                              start_uid = blast_uid,
                              end_uid = query_uid,
                              properties = { "subject": False, "start": query_start, "end": query_end } )
        
        endpoint.create_edge( label = schema.BlastQueriesSequence,
                              start_uid = blast_uid,
                              end_uid = subject_uid,
                              properties = { "subject": True, "start": subject_start, "end": subject_end } )
        
        # Expand TARGET edge to TARGET node
        query_ss_name = "{0}({1},{2})".format( query_uid, query_start, query_end )
        
        endpoint.create_node( label = schema.Subsequence,
                              uid = query_ss_name,
                              properties = { "start": query_start, "end": query_end } )
        
        endpoint.create_edge( label = schema.BlastUsesSubsequence,
                              start_uid = blast_uid,
                              end_uid = query_ss_name,
                              properties = { "is_subject": False } )
        
        endpoint.create_edge( label = schema.SequenceContainsSubsequence,
                              start_uid = query_uid,
                              end_uid = query_ss_name,
                              properties = { } )
        
        subject_ss_name = "{0}({1},{2})".format( subject_uid, subject_start, subject_end )
        
        endpoint.create_node( label = schema.Subsequence,
                              uid = subject_ss_name,
                              properties = { "start": subject_start, "end": subject_end } )
        
        endpoint.create_edge( label = schema.BlastUsesSubsequence,
                              start_uid = blast_uid,
                              end_uid = subject_ss_name,
                              properties = { "is_subject": True } )
        
        endpoint.create_edge( label = schema.SequenceContainsSubsequence,
                              start_uid = subject_uid,
                              end_uid = subject_ss_name,
                              properties = { } )
        
        endpoint.create_edge( label = schema.SubsequenceBlastqueriedSubsequence,
                              start_uid = query_ss_name,
                              end_uid = subject_ss_name,
                              properties = properties )


unknown = str

# noinspection SpellCheckingInspection
BLASTMAP = {
    # BLAST-name    Description                                 DB name                       Type      Special
    "qseqid"     : ("Query Seq-id", "query_accession", str, _SPC_QUERY_UID),
    "qgi"        : ("Query GI", "query_gi", unknown, _SPC_NONE_),
    "qacc"       : ("Query accesion", "query_accession", unknown, _SPC_NONE_),
    "sseqid"     : ("Subject Seq-id", "subject_accession", str, _SPC_SUBJECT_UID),
    "sallseqid"  : ("All subject Seq-id(s)", "all_subject_accessions", unknown, _SPC_NONE_),
    "sgi"        : ("Subject GI", "subject_gi", unknown, _SPC_NONE_),
    "sallgi"     : ("All subject GIs", "all_subject_gis", unknown, _SPC_NONE_),
    "sacc"       : ("Subject accession", "subject_accession", unknown, _SPC_NONE_),
    "sallacc"    : ("All subject accessions", "all_subject_accessions", unknown, _SPC_NONE_),
    "qstart"     : ("Start of alignment in query", "query_start", int, _SPC_QUERY_START),
    "qend"       : ("End of alignment in query", "query_end", int, _SPC_QUERY_END),
    "sstart"     : ("Start of alignment in subject", "subject_start", int, _SPC_SUBJECT_START),
    "send"       : ("End of alignment in subject", "subject_end", int, _SPC_SUBJECT_END),
    "qseq"       : ("Aligned part of query sequence", "aligned_query_sequence", unknown, _SPC_NONE_),
    "sseq"       : ("Aligned part of subject sequence", "aligned_subject_sequence", unknown, _SPC_NONE_),
    "evalue"     : ("Expect value", "e_value", float, _SPC_NONE_),
    "bitscore"   : ("Bit score", "bit_score", float, _SPC_NONE_),
    "score"      : ("Raw score", "raw_score", unknown, _SPC_NONE_),
    "length"     : ("Alignment length", "alignment_length", int, _SPC_NONE_),
    "pident"     : ("Percentage of identical matches", "percent_identity", float, _SPC_NONE_),
    "nident"     : ("Number of identical matches", "num_identical", int, _SPC_NONE_),
    "mismatch"   : ("Number of mismatches", "mismatches", int, _SPC_NONE_),
    "positive"   : ("Number of positive-scoring matches", "num_positive", int, _SPC_NONE_),
    "gapopen"    : ("Number of gap openings", "gap_opens", int, _SPC_NONE_),
    "gaps"       : ("Total number of gap", "num_gaps", int, _SPC_NONE_),
    "slen"       : ("Subject length (?)", "subject_length", int, _SPC_NONE_),
    "qlen"       : ("Query length (?)", "query_length", int, _SPC_NONE_),
    "ppos"       : ("Percentage of positive-scoring matches", "percent_positive", float, _SPC_NONE_),
    "frames"     : ("Query and subject frames", "frames", unknown, _SPC_NONE_),
    "qframe"     : ("Query frame", "query_frame", unknown, _SPC_NONE_),
    "sframe"     : ("Subject frame", "subject_frame", unknown, _SPC_NONE_),
    "btop"       : ("Blast traceback operations", "traceback", unknown, _SPC_NONE_),
    "staxids"    : ("unique Subject Taxonomy ID(s)", "subject_taxonomy", unknown, _SPC_NONE_),
    "sscinames"  : ("unique Subject Scientific Name(s)", "subject_scientific_name", unknown, _SPC_NONE_),
    "scomnames"  : ("unique Subject Common Name(s)", "subject_common_name", unknown, _SPC_NONE_),
    "sblastnames": ("unique Subject Blast Name(s)", "subject_blast_name", unknown, _SPC_NONE_),
    "sskingdoms" : ("unique Subject Super Kingdom(s)", "subject_super_kingdom", unknown, _SPC_NONE_),
    "stitle"     : ("Subject Title", "subject_title", unknown, _SPC_NONE_),
    "salltitles" : ("All Subject Title(s)", "all_subject_titles", unknown, _SPC_NONE_),
    "sstrand"    : ("Subject Strand", "subject_strand", unknown, _SPC_NONE_),
    "qcovs"      : ("Query Coverage Per Subject", "query_coverage_per_subject", unknown, _SPC_NONE_),
    "qcovhsp"    : ("Query Coverage Per HSP", "query_coverage_per_hsp", unknown, _SPC_NONE_),
    "qcovus"     : ("Measure of Query Coverage", "query_coverage", unknown, _SPC_NONE_)
}

# noinspection SpellCheckingInspection
BLASTAUTOMAP = {
    # ID    Description                                                 Translation
    "std": ("BLAST standard", ("qseqid", "sseqid", "pident", "length", "mismatch", "gapopen", "qstart", "qend", "sstart", "send", "evalue", "bitscore")),
    "pcn": ("J. Pathmanathan's clean network for Composite Searcher", ("qseqid", "sseqid", "evalue", "pident", "bitscore", "qstart", "qend", "qlen", "sstart", "send", "slen"))
}


def __blast_columns_help():
    """
    Lists the BLAST column mapping, taken from the BLAST+ documentation at https://www.ncbi.nlm.nih.gov/books/NBK279675/.
    """
    r = []
    
    r.append( "Details on BLAST columns" )
    r.append( "Columns" )
    r.append( "<table>" )
    r.append( intermake.pr.fmt_row( ["BLAST+ name", "Description", intermake.Controller.ACTIVE.app.name + " name (once in database)"] ) )
    
    for k, v in BLASTMAP.items():
        description = v[0]
        name = v[1]  # name
        
        if v[3] != _SPC_NONE_:  # special
            description += " **MANDATORY**"
        
        r.append( intermake.pr.fmt_row( [k, description, name] ) )
    
    r.append( "</table>" )
    
    r.append( "Special identifiers" )
    r.append( "<table>" )
    r.append( intermake.pr.fmt_row( ["ID", "Description", "Translation"] ) )
    
    for k, v in BLASTAUTOMAP.items():
        r.append( intermake.pr.fmt_row( [k, v[0], ", ".join( v[1] )] ) )
    
    r.append( "</table>" )
    
    return "\n".join( r )


intermake.Application.LAST.help.add( "blast_columns", "Quick reference sheet for the available columns in BLAST", __blast_columns_help )



"""
BIO42_GUI isn't a application, it's a library that just exports a
`GuiController`: `B42GuiController`.

This is used by BIO42 when the user runs the `GUI` command.
"""

from bio42_gui.controller.controller import B42GuiController



"""
Includes a set of default scripts for B42.

These are loaded when starting `FrmQueryBrowser` in the intermake.GUI, or may be loaded, in
intermake.CLI or Python proper, by using `import bio42_scrips`.

Note that the scripts are registered with the application - they are not "exported" as Python
variables (though from intermake.PYI they *will* be available as defined locals).
"""

from bio42_scripts import scripts as _scripts



"""
BIO42 API.
"""

__version__ = "0.0.0.27"

from .application import Application
from .commands import download_accessions, download_annotations, download_taxonomy, download_ontology, import_annotations, import_blast, import_csv, import_go, import_taxonomy, import_sequences, ESequenceFlags

#
# We also include some the Intermake/Neocommand API for convenience
#
from intermake.commands import print_version, print_help
from neocommand import PRIMARY_KEY, NULL_ENDPOINT, ECHOING_ENDPOINT, set_database, save_script, send_cypher, control_server, ENeo4jCommand, Neo4jStatus, test_connection, apply_schema, open_parcel, open_database, close_endpoint, open_file, EOperatingSystem, transfer, EParcelMethod, print_endpoints
from mhelper import ignore as _ignore


_ignore( print_version, PRIMARY_KEY, NULL_ENDPOINT, ECHOING_ENDPOINT, set_database, save_script, send_cypher, control_server, ENeo4jCommand, Neo4jStatus, test_connection, apply_schema, open_parcel, open_database, close_endpoint, open_file, EOperatingSystem, transfer, EParcelMethod, print_help, print_endpoints )



"""
Subpackage containing the commands associated with importing data.
"""

from .parser_annotations import import_annotations
from .parser_blast import import_blast
from .parser_csv import import_csv
from .parser_go import import_go
from .parser_taxonomy import import_taxonomy
from .parser_sequence import import_sequences, ESequenceFlags




from itertools import chain
from os import path
from typing import Optional, cast, List
from PyQt5.QtWidgets import QTreeWidget, QGridLayout, QTreeWidgetItem
from bio42 import Application

from mhelper import ArgInspector, ArgValueCollection, Sentinel, SwitchError, ansi_format_helper, file_helper, ignore, io_helper

import re
import editorium as ed
import intermake
import intermake_qt
import mgraph as mg
import neocommand as nc
import mhelper_qt as qt

from bio42_gui.forms.designer import frm_query_browser_designer
from bio42_gui.forms.designer.resources import resources
from bio42_gui.forms.frm_b42_base import AbstractB42SubWindow
from bio42_gui.controller.controller import B42GuiController


CUSTOM_QUERY = Sentinel( "(Custom query)" )
GUI_EP = Sentinel( "(GUI)" )


class GuiEndpoint( nc.MGraphEndpoint ):
    def __init__( self ):
        super().__init__( "GuiEndpoint" )


@Application.INSTANCE.command( visibility = intermake.visibilities.INTERNAL )
def _load_scripts() -> None:
    # Ensure Scripts are loaded
    import bio42_scripts
    ignore( bio42_scripts )


class FrmQueryBrowser( AbstractB42SubWindow ):
    """
    Allows the user to execute and view Cypher queries.
    """
    
    
    def __init__( self, parent: qt.QWidget ):
        """
        CONSTRUCTOR
        """
        super().__init__( parent )
        self.ui = frm_query_browser_designer.Ui_Dialog( self )
        self.setWindowTitle( "Query" )
        self.ui.PAGER_SET.setCurrentIndex( 0 )
        
        # Remove designing styles
        self.ui.FRA_TREE_HOLDER.setStyleSheet( "" )
        self.ui.FRA_BROWSER_HOLDER.setStyleSheet( "" )
        
        # Queries (populated later)
        self.queries = qt.ComboBoxWrapper( self.ui.CMB_QUERY, [] )
        self.queries.enabled = False
        
        # Browser
        self.browser_ctrl = None
        self.browser_page = None
        self.is_browser = False
        
        if B42GuiController.get_settings().enable_browser:
            self.enable_browser()
        
        # Received status
        self.received_html = ""
        self.received_map = { }
        self.received_root = None
        
        # Sent status
        self.editing_script: nc.ScriptCommand = None
        
        # Arguments
        self.editorium = ed.EditoriumGrid( self.ui.LAY_PARAMETERS, self.main_window.controller.editorium )
        self.editorium.fn_arg = self.__editorium_fn_arg
        
        # Signals
        self.ignore_changes = False
        self.ui.CMB_QUERY.currentIndexChanged[int].connect( self.__on_CMB_QUERY_currentIndexChanged )
        self.ui.TXT_CYPHER.textChanged.connect( self.__on_TXT_CYPHER_textChanged )
        
        # Tree view
        lay = QGridLayout()
        self.ui.FRA_TREE_HOLDER.setLayout( lay )
        self.results_tree = QTreeWidget()
        lay.addWidget( self.results_tree )
        
        # Sources
        self.sources = qt.ComboBoxWrapper( self.ui.CMB_SOURCE, (x for x in nc.get_core().endpoint_manager if isinstance( x, nc.DatabaseEndpoint )) )
        
        if len( self.sources ) == 0:
            qt.QMessageBox.warning( parent, "Oops", "You haven't defined any database endpoints. You won't be able to issue a query until you do so." )
        
        # Destinations
        self.destinations = qt.ComboBoxWrapper( self.ui.CMB_DESTINATION,
                                                chain( [GUI_EP],
                                                       (x for x in nc.get_core().endpoint_manager if isinstance( x, nc.Destination )) ) )
        
        # Browser?
        self.__scripts_loaded = False
        self.ui.CHK_ENABLE_BROWSER.setChecked( self.gui.get_settings().enable_browser )
        self.ui.CHK_ENABLE_BROWSER.stateChanged[int].connect( self.__on_CHK_ENABLE_BROWSER_stateChanged )
        self.ui.CHK_B42_SCRIPTS.setChecked( self.gui.b42_gui_settings.auto_load_bio42_scripts )
        self.ui.CHK_B42_SCRIPTS.stateChanged[int].connect( self.__on_CHK_B42_SCRIPTS_stateChanged )
    
    
    def btn_style( self, button: qt.QToolButton ):
        button.setFixedSize( qt.QSize( qt.QWIDGETSIZE_MAX, qt.QWIDGETSIZE_MAX ) )
        button.setIconSize( qt.QSize( 16, 16 ) )
        button.setToolButtonStyle( qt.Qt.ToolButtonFollowStyle )
        button.setProperty( "style", "sidearea" )
        button.setStyleSheet( "" )
    
    
    @property
    def gui( self ) -> B42GuiController:
        return cast( B42GuiController, intermake.Controller.ACTIVE )
    
    
    def __on_CHK_ENABLE_BROWSER_stateChanged( self, _ ):
        self.gui.get_settings().enable_browser = self.ui.CHK_ENABLE_BROWSER.isChecked()
    
    
    def __on_CHK_B42_SCRIPTS_stateChanged( self, _ ):
        self.gui.b42_gui_settings.auto_load_bio42_scripts = self.ui.CHK_B42_SCRIPTS.isChecked()
        
        if self.ui.CHK_B42_SCRIPTS.isChecked():
            self.__request_load_scripts_library()
    
    
    def on_shown( self ):
        if self.gui.b42_gui_settings.auto_load_bio42_scripts:
            self.__request_load_scripts_library()
        else:
            self.populate_scripts()
    
    
    def __request_load_scripts_library( self ):
        if not self.__scripts_loaded:
            def ___callback( result: intermake.Result ):
                assert result.command is intermake.BasicCommand.retrieve( _load_scripts ), result.command
                self.populate_scripts()
            
            
            self.__scripts_loaded = True
            intermake.Controller.ACTIVE.acquire( command = _load_scripts, window = self ).run().listen( ___callback )
    
    
    def apply_view( self, html = "", node_map = None, root: intermake.Result = None ):
        self.received_html = html
        self.received_map = node_map
        self.received_root = root
        
        # Tree
        if self.received_root is not None and self.received_root.is_success:
            self.__set_results_view_on( self.received_root.result )
        else:
            self.__set_results_view_on( None )
        
        # HTML
        if self.is_browser:
            from PyQt5.QtWebEngineWidgets import QWebEnginePage
            wep: QWebEnginePage = self.browser_page
            file_name = self.write_html_to_file()
            wep.load( qt.QUrl.fromLocalFile( file_name ) )  # nb. setHtml doesn't work with visjs, so we always need to use a temporary file
    
    
    def list_list_queries( self ):
        return file_helper.list_dir( self.get_query_folder() )
    
    
    def populate_scripts( self ):
        self.queries = qt.ComboBoxWrapper( self.ui.CMB_QUERY,
                                           tuple( chain( [CUSTOM_QUERY],
                                                         self.list_list_queries(),
                                                         (x for x in intermake.Controller.ACTIVE.app.commands if isinstance( x, nc.ScriptCommand )) ) ) )
        
        self.queries.enabled = True
        self.queries.selected = CUSTOM_QUERY
    
    
    def __editorium_fn_arg( self, arg: ArgInspector ) -> Optional[ArgInspector]:
        if arg.name in (nc.ScriptCommand.ARG_DATABASE,
                        nc.ScriptCommand.ARG_OUTPUT,
                        nc.ScriptCommand.ARG_QUIET,
                        nc.ScriptCommand.ARG_RETRIES,
                        nc.ScriptCommand.ARG_TIMEOUT,
                        nc.ScriptCommand.ARG_DUMP):
            return None
        
        return arg
    
    
    def __on_TXT_CYPHER_textChanged( self ):
        if self.ignore_changes:
            return
        
        self.ignore_changes = True
        self.ui.CMB_QUERY.setCurrentIndex( 0 )
        self.ignore_changes = False
        self.__on_CMB_QUERY_currentIndexChanged( 0 )
    
    
    def __on_CMB_QUERY_currentIndexChanged( self, _ ):
        if self.ignore_changes:
            return
        
        # Find the script!
        q: object = self.queries.selected
        
        # Textbox
        self.ignore_changes = True
        
        if q is CUSTOM_QUERY:
            self.set_script( nc.ScriptCommand( cypher = self.ui.TXT_CYPHER.toPlainText(),
                                               name = "custom_query",
                                               filename = self.editing_script.filename if self.editing_script is not None else None ),
                             update_text = False )
        elif isinstance( q, str ):
            self.load_file( q )
        elif isinstance( q, nc.ScriptCommand ):
            self.set_script( q )
        elif q is not None:
            raise SwitchError( "self.queries.selected", q, instance = True )
        
        self.ignore_changes = False
        
        # Arguments
        if isinstance( self.editing_script, nc.ScriptCommand ):
            self.editing_args = ArgValueCollection( self.editing_script.args )
        else:
            self.editing_args = None
        
        self.editorium.target = self.editing_args
        self.editorium.recreate()
    
    
    def __format_node( self, node: mg.MNode ):
        ent: nc.NcNode = node.data
        return nc.get_core().schema.describe( ent )
    
    
    def write_html_to_file( self ):
        file_name = path.join( intermake.Controller.ACTIVE.app.local_data.local_folder( intermake.constants.FOLDER_TEMPORARY ), "temp.html" )
        file_helper.write_all_text( file_name, self.received_html )
        return file_name
    
    
    def enable_browser( self ):
        from PyQt5.QtWebEngineWidgets import QWebEngineView, QWebEnginePage
        
        owner = self
        
        
        class MyPage( QWebEnginePage ):
            def javaScriptConsoleMessage( self, level, message: str, lineNumber: int, source_id: str ):
                owner.handle_web_interaction( message )
        
        
        self.browser_ctrl = QWebEngineView()
        self.browser_ctrl.setVisible( True )
        self.browser_page = MyPage()
        self.browser_ctrl.setPage( self.browser_page )
        self.ui.GRID_BROWSER_HOLDER.addWidget( self.browser_ctrl )
        self.is_browser = True
    
    
    def handle_web_interaction( self, message ):
        """
        Handles messages RECEIVED FROM the web-browser. 
        :param message: 
        :return: 
        """
        
        # MGRAPH sends messages always start with "MGRAPH". 
        if not message.startswith( "MGRAPH " ):
            return
        
        # The command is of the form MGRAPH command ( args )
        m1 = re.search( "MGRAPH ([a-zA-Z_]+) *\((.+)\)", message )
        
        if not m1:
            return
        
        command = m1.group( 1 )
        args = m1.group( 2 ).strip()
        
        if command == "SELECT_NODE":
            # A node or nodes has been selected
            node_ids = [int( x.strip() ) for x in args.split( "," )]
            
            if len( node_ids ) != 1:
                return
            
            # NeoCommand's created MGraph stores the actual node in `x~~MNode::data~~NodeTag::entity~~MNode`
            node_id = node_ids[0]
            node: mg.MNode = self.received_map.get( node_id )
            
            if node is None:
                return
            
            self.__set_results_view_on( node.data )
    
    
    @qt.exqtSlot()
    def on_BTN_SYSTEM_BROWSER_clicked( self ) -> None:
        """
        Signal handler:
        """
        file_name = self.write_html_to_file()
        io_helper.system_open( file_name )  # `webbrowser` is for URLs, use `open` instead.
    
    
    @qt.exqtSlot()
    def on_BTN_SEND_TO_clicked( self ) -> None:
        """
        Signal handler:
        """
        if not isinstance( self.received_root, nc.Origin ):
            qt.QMessageBox.warning( self, "Oops", "The current data ({}) isn't exportable because it can't act as a data-source. Make sure you have received some data and that you have selected an exportable destination for your query.".format( self.received_root ) )
            return
        
        from .frm_transfer import FrmTransfer
        FrmTransfer.request( self, source = self.received_root )
    
    
    @qt.exqtSlot()
    def on_BTN_EXPORT_AS_clicked( self ) -> None:
        """
        Signal handler:
        """
        if not isinstance( self.received_root, GuiEndpoint ):
            qt.QMessageBox.warning( self,
                                    "Oops",
                                    "The current data ({}) isn't a graph. "
                                    "Make sure that you have selected a graph destination for your query and try again."
                                    .format( self.received_root ) )
            return
        
        graph = self.received_root.read_graph()
        
        if len( graph.nodes ) == 0:
            qt.QMessageBox.warning( self,
                                    "Oops",
                                    "The current graph isn't exportable because it doesn't contain any nodes. "
                                    "Make sure you have received some graph data first." )
        
        file_name = qt.qt_gui_helper.browse_save( self, ";;".join( "{} (*{})".format( x[1][0], x[0] ) for x in mg.exporting.EXTENSION_MAP.items() ) )
        
        if file_name is not None:
            mg.exporting.export_file( graph,
                                      file_name,
                                      fprops = lambda x: cast( x.data, mg.MNode ).properties )
    
    
    @qt.exqtSlot()
    def on_BTN_CUSTOM_LIBRARIES_clicked( self ) -> None:
        """
        Signal handler:
        """
        self.show_command( intermake.import_python_module )
    
    
    @qt.exqtSlot()
    def on_BTN_SAVE_clicked( self ) -> None:
        """
        Signal handler:
        """
        save_over_current = "Save"
        save_to_list = "Save to list..."
        remove_from_list = "Delete"
        save_to_file = "Save to file..."
        opts = [save_over_current, save_to_list, remove_from_list, save_to_file]
        
        sel = qt.menu_helper.show_menu( self, opts )
        
        file_name = self.editing_script.filename if self.editing_script is not None else None
        
        if sel is save_over_current:
            if not file_name:
                qt.QMessageBox.information( self,
                                            self.windowTitle(),
                                            "You must “<i>save to list...</i>” or “<i>save as...</i>” before you can use “<i>save</i>”." )
                return
            
            self.save_file( file_name )
        
        if sel is save_to_list:
            if not self.editing_script:
                qt.QMessageBox.information( self,
                                            self.windowTitle(),
                                            "You must “<i>save to list...</i>” or “<i>save as...</i>” before you can use “<i>save</i>”." )
                return
            
            query_name, ok = qt.QInputDialog.getText( self, "Save query", "Please name your query" )
            
            if ok:
                self.save_file( self.list_name_to_file_name( query_name ) )
        elif sel is remove_from_list:
            if not file_name:
                qt.QMessageBox.information( self,
                                            self.windowTitle(),
                                            "This script is not associated with a file that can be deleted." )
                return
            
            if qt.QMessageBox.question( self,
                                        self.windowTitle(),
                                        "Are you sure that you wish to delete this file?\n{}".format( self.editing_script.filename ),
                                        qt.QMessageBox.Ok | qt.QMessageBox.Cancel
                                        ) != qt.QMessageBox.Ok:
                return
            
            file_helper.recycle_file( self.editing_script.filename )
            self.editing_script.filename = None
        elif sel is save_to_file:
            file_name, _ = qt.QFileDialog.getSaveFileName( self, "Save query", filter = "Queries (*.cypher)" )
            
            if file_name:
                self.save_file( file_name )
        elif sel is not None:
            raise SwitchError( "sel", sel )
    
    
    def save_file( self, file_name: str ):
        if isinstance( file_name, str ):
            self.editing_script = nc.ScriptCommand( cypher = self.ui.TXT_CYPHER.toPlainText(), 
                                                    filename = file_name )
            file_helper.write_all_text( file_name, self.ui.TXT_CYPHER.toPlainText() )
            self.update_title()
        else:
            raise ValueError( "Cannot save - no filename." )
    
    
    def load_file( self, file_name: str ):
        script = nc.ScriptCommand( cypher = file_helper.read_all_text( file_name ),
                                   name = file_helper.get_filename_without_extension( file_name ), 
                                   filename = file_name )
        self.set_script( script )
    
    
    def set_script( self, script: nc.ScriptCommand, update_text: bool = True ):
        if update_text:
            self.ui.TXT_CYPHER.setPlainText( script.plain_text )
        self.editing_script = script
        self.update_title()
    
    
    def update_title( self ):
        self.setWindowTitle( "Query - {} - {}".format( self.editing_script.name, self.editing_script.filename or "(no_file)" ) )
    
    
    def list_name_to_file_name( self, n ):
        return path.join( self.get_query_folder(), n + ".cypher" )
    
    
    def get_query_folder( self ):
        return intermake.Controller.ACTIVE.app.local_data.local_folder( "queries" )
    
    
    @qt.exqtSlot()
    def on_BTN_LOAD_clicked( self ) -> None:
        """
        Signal handler:
        """
        file_name, _ = qt.QFileDialog.getOpenFileName( self, "Open query", filter = "Queries (*.cypher)" )
        
        if file_name:
            self.load_file( file_name )
    
    
    def on_BTN_SHOW_ADVANCED_clicked( self ) -> None:
        """
        Signal handler:
        """
        self.ui.PAGER_SET.setCurrentIndex( 1 )
    
    
    @qt.exqtSlot()
    def on_BTN_BACK_TO_QUERY_clicked( self ) -> None:
        """
        Signal handler:
        """
        self.ui.PAGER_SET.setCurrentIndex( 0 )
    
    
    @qt.exqtSlot()
    def on_BTN_LOCAL_BROWSER_clicked( self ) -> None:
        """
        Signal handler:
        """
        # TODO: these resources placeholders only and are wrong
        self.ui.FRA_BROWSER_HOLDER_BASE.setVisible( not self.ui.FRA_BROWSER_HOLDER_BASE.isVisible() )
        self.ui.BTN_LOCAL_BROWSER.setIcon( resources.next.icon() if self.ui.FRA_BROWSER_HOLDER_BASE.isVisible() else resources.view_list.icon() )
    
    
    @qt.exqtSlot()
    def on_BTN_TREE_clicked( self ) -> None:
        """
        Signal handler:
        """
        self.ui.FRA_TREE_HOLDER_BASE.setVisible( not self.ui.FRA_TREE_HOLDER_BASE.isVisible() )
        self.ui.BTN_TREE.setIcon( resources.next.icon() if self.ui.FRA_TREE_HOLDER_BASE.isVisible() else intermake_qt.view_list.expanddown.icon() )
    
    
    @qt.exqtSlot()
    def on_BTN_EXECUTE_clicked( self ) -> None:
        """
        Signal handler:
        """
        
        # Database
        database: nc.DatabaseEndpoint = self.sources.selected
        
        if database is None:
            qt.QMessageBox.warning( self, "Oops", "A source must be specified." )
            return
        
        # Destination
        destination = self.destinations.selected
        
        if destination is GUI_EP:
            destination = GuiEndpoint()
        elif destination is None:
            qt.QMessageBox.warning( self, "Oops", "A destination must be specified." )
            return
        
        self.apply_view( html = "<html><body>Please wait...</body></html>" )
        
        self.editing_args.set_value( nc.ScriptCommand.ARG_DATABASE, database )
        self.editing_args.set_value( nc.ScriptCommand.ARG_OUTPUT, destination )
        
        intermake.acquire( command = self.editing_script,
                           window = self,
                           auto_close = True,
                           ).run( **self.editing_args.tokwargs() ).listen( self.__execute_complete )
    
    
    def __execute_complete( self, result: intermake.Result ):
        if result.is_error:
            # Display errors directly
            self.apply_view( html = qt.qt_gui_helper.ansi_to_html( ansi_format_helper.format_traceback( result.exception, result.traceback ) ),
                             root = result )
        elif isinstance( result.result, nc.DbStats ) and isinstance( result.result.endpoint, GuiEndpoint ):
            # Result as a `GuiEndpoint`
            if len( result.result.endpoint.graph.nodes ) != 0:
                map = { }
                self.apply_view( html = mg.exporting.export_vis_js( result.result.endpoint.graph,
                                                                    fnode = self.__format_node,
                                                                    rooted = False,
                                                                    map = map ),
                                 root = result,
                                 node_map = map )
            else:
                self.apply_view( html = self.__extra_data_to_html( result.result.endpoint.extra_data ),
                                 root = result )
        else:
            self.apply_view( html = str( result.result ),
                             root = result )
    
    
    def __extra_data_to_html( self, items: List[object] ):
        r = []
        r.append( "<html><title>Result of query</title><body>" )
        
        r.append( "Query result does not contain a graph." )
        
        r.append( "<ul>" )
        
        if len( items ) == 0:
            r.append( "<li>No items</li>" )
        
        for item in items:
            r.append( "<li>" )
            
            if isinstance( item, nc.NcData ):
                r.append( "{} = {}".format( item.varname, item.value ) )
            else:
                r.append( "{}".format( item ) )
            
            r.append( "</li>" )
        
        r.append( "</ul></body></html>" )
        
        return "\n".join( r )
    
    
    def __set_results_view_on( self, x ):
        self.results_tree.clear()
        
        self.results_tree.setHeaderItem( QTreeWidgetItem() )
        self.results_tree.headerItem().setText( 0, "name" )
        self.results_tree.headerItem().setText( 1, "type" )
        self.results_tree.headerItem().setText( 2, "value" )
        
        if isinstance( x, nc.NcEntity ):
            self.results_tree.addTopLevelItem( self.__make_tree_node( x.varname, x ) )
        else:
            self.results_tree.addTopLevelItem( self.__make_tree_node( "untitled", x ) )
    
    
    def __make_tree_node( self, n, x ):
        item = QTreeWidgetItem()
        
        if isinstance( x, mg.MNode ):
            x = x.data
        elif isinstance( x, mg.MEdge ):
            x = x.data
        
        if isinstance( x, nc.DbStats ):
            item.setText( 0, n )
            item.setText( 1, "Result" )
            item.setText( 2, str( x ) )
            
            for k, v in x.__dict__.items():
                if not k.startswith( "_" ):
                    item.addChild( self.__make_tree_node( ".{}".format( k ), v ) )
        elif isinstance( x, intermake.Result ):
            item.setText( 0, n )
            item.setText( 1, "Result" )
            item.setText( 2, x.command )
            item.addChild( self.__make_tree_node( "value", x.result ) )
        elif isinstance( x, GuiEndpoint ):
            item.setText( 0, n )
            item.setText( 1, "Endpoint" )
            item.setText( 2, str( x ) )
            item.addChild( self.__make_tree_node( "value", x.graph ) )
        elif isinstance( x, mg.MGraph ):
            item.setText( 0, n )
            item.setText( 1, "Graph" )
            item.setText( 2, str( x ) )
            
            for i, v in enumerate( x.nodes ):
                item.addChild( self.__make_tree_node( "node{}".format( i ), v ) )
            
            for i, v in enumerate( x.edges ):
                item.addChild( self.__make_tree_node( "edge{}".format( i ), v ) )
        elif isinstance( x, nc.NcNode ):
            item.setText( 0, n )
            item.setText( 1, "Node" )
            item.setText( 2, x.label )
            
            for k, v in x.properties.items():
                item.addChild( self.__make_tree_node( ".{}".format( k ), v ) )
        elif isinstance( x, nc.NcEdge ):
            item.setText( 0, n )
            item.setText( 1, "Edge" )
            item.setText( 2, x.label )
            
            item.addChild( self.__make_tree_node( "start", x.start ) )
            item.addChild( self.__make_tree_node( "end", x.end ) )
            
            for k, v in x.properties.items():
                item.addChild( self.__make_tree_node( ".{}".format( k ), v ) )
        elif isinstance( x, nc.NcData ):
            item.setText( 0, n or x.varname )
            item.setText( 1, "Data" )
            item.setText( 2, str( x.value ) )
        else:
            item.setText( 0, n )
            item.setText( 1, "Object" )
            item.setText( 2, str( x ) )
        
        return item



"""
Neo4j doesn't have a well defined 'schema', this class just tells the underlying
library (neocommand) what our Bio42 database is supposed to look like so we
can check it against this and we have somewhere to formally define the names of
everything.
"""

from neocommand import NodeSchema as _node, EdgeSchema as _edge, PropertySchema as _data, DatabaseSchema
from typing import TypeVar


_T = TypeVar( "_T" )
available = []


def register( schema: DatabaseSchema ):
    for x in available:
        schema.register( x )


def _add( x: _T ) -> _T:
    available.append( x )
    return x


# noinspection SpellCheckingInspection
Blast = _add( _node( label = "Blast",
                     is_complete = True,
                     properties = (_data( "query_accession", str, doc = "Query Seq-id, blast name = qseqid, qacc" ),
                                   _data( "query_gi", str, doc = "Query GI, blast name = 'qgi'" ),
                                   _data( "all_subject_accessions", str, doc = "All subject Seq-id(s), blast name = 'sallseqid, sallacc'" ),
                                   _data( "subject_gi", str, doc = "Subject GI, blast name = 'sgi'" ),
                                   _data( "all_subject_gis", str, doc = "All subject GIs, blast name = 'sallgi'" ),
                                   _data( "subject_accession", str, doc = "Subject accession, blast name = 'sseqid, sacc'" ),
                                   _data( "query_start", int, doc = "Start of alignment in query, blast name = 'qstart'" ),
                                   _data( "query_end", int, doc = "End of alignment in query, blast name = 'qend'" ),
                                   _data( "subject_start", int, doc = "Start of alignment in subject, blast name = 'sstart'" ),
                                   _data( "subject_end", int, doc = "End of alignment in subject, blast name = 'send'" ),
                                   _data( "aligned_query_sequence", str, doc = "Aligned part of query sequence, blast name = 'qseq'" ),
                                   _data( "aligned_subject_sequence", str, doc = "Aligned part of subject sequence, blast name = 'sseq'" ),
                                   _data( "e_value", float, doc = "Expect value, blast name = 'evalue'" ),
                                   _data( "bit_score", float, doc = "Bit score, blast name = 'bitscore'" ),
                                   _data( "raw_score", str, doc = "Raw score, blast name = 'score'" ),
                                   _data( "alignment_length", int, doc = "Alignment length, blast name = 'length'" ),
                                   _data( "percent_identity", float, doc = "Percentage of identical matches, blast name = 'pident'" ),
                                   _data( "num_identical", int, doc = "Number of identical matches, blast name = 'nident'" ),
                                   _data( "mismatches", int, doc = "Number of mismatches, blast name = 'mismatch'" ),
                                   _data( "num_positive", int, doc = "Number of positive-scoring matches, blast name = 'positive'" ),
                                   _data( "gap_opens", int, doc = "Number of gap openings, blast name = 'gapopen'" ),
                                   _data( "num_gaps", int, doc = "Total number of gap, blast name = 'gaps'" ),
                                   _data( "subject_length", int, doc = "Subject length (?), blast name = 'slen'" ),
                                   _data( "query_length", int, doc = "Query length (?), blast name = 'qlen'" ),
                                   _data( "percent_positive", float, doc = "Percentage of positive-scoring matches, blast name = 'ppos'" ),
                                   _data( "frames", str, doc = "Query and subject frames, blast name = 'frames'" ),
                                   _data( "query_frame", str, doc = "Query frame, blast name = 'qframe'" ),
                                   _data( "subject_frame", str, doc = "Subject frame, blast name = 'sframe'" ),
                                   _data( "traceback", str, doc = "Blast traceback operations, blast name = 'btop'" ),
                                   _data( "subject_taxonomy", str, doc = "unique Subject Taxonomy ID(s), blast name = 'staxids'" ),
                                   _data( "subject_scientific_name", str, doc = "unique Subject Scientific Name(s), blast name = 'sscinames'" ),
                                   _data( "subject_common_name", str, doc = "unique Subject Common Name(s), blast name = 'scomnames'" ),
                                   _data( "subject_blast_name", str, doc = "unique Subject Blast Name(s), blast name = 'sblastnames'" ),
                                   _data( "subject_super_kingdom", str, doc = "unique Subject Super Kingdom(s), blast name = 'sskingdoms'" ),
                                   _data( "subject_title", str, doc = "Subject Title, blast name = 'stitle'" ),
                                   _data( "all_subject_titles", str, doc = "All Subject Title(s), blast name = 'salltitles'" ),
                                   _data( "subject_strand", str, doc = "Subject Strand, blast name = 'sstrand'" ),
                                   _data( "query_coverage_per_subject", str, doc = "Query Coverage Per Subject, blast name = 'qcovs'" ),
                                   _data( "query_coverage_per_hsp", str, doc = "Query Coverage Per HSP, blast name = 'qcovhsp'" ),
                                   _data( "query_coverage", str, doc = "Measure of Query Coverage, blast name = 'qcovus'" )) ) )

Family = _add( _node( "Family" ) )

Feature = _add( _node( "Feature" ) )

Reference = _add( _node( "Reference" ) )

Source = _add( _node( "Source" ) )

Plasmid = _add( _node( "Plasmid" ) )

Taxon = _add( _node( label = "Taxon",
                     is_complete = True,
                     properties = (_data( "parent_id", str, key = "Taxon" ),
                                   _data( "rank", str ),
                                   _data( "embl_code", str ),
                                   _data( "division_id", int ),
                                   _data( "inherited_div_flag", bool ),
                                   _data( "genetic_code_id", int ),
                                   _data( "inherited_gc_flag", bool ),
                                   _data( "mitochondrial_genetic_code_id", int ),
                                   _data( "inherited_mgc_flag", bool ),
                                   _data( "genbank_hidden_flag", bool ),
                                   _data( "hidden_subtree_root_flag", bool ),
                                   _data( "comments", str ),
                                   _data( "acronym", str ),
                                   _data( "anamorph", str ),
                                   _data( "authority", str ),
                                   _data( "blast_name", str ),
                                   _data( "common_name", str ),
                                   _data( "equivalent_name", str ),
                                   _data( "genbank_acronym", str ),
                                   _data( "genbank_anamorph", str ),
                                   _data( "genbank_common_name", str ),
                                   _data( "genbank_synonym", str ),
                                   _data( "in_part", str ),
                                   _data( "includes", str ),
                                   _data( "misnomer", str ),
                                   _data( "misspelling", str ),
                                   _data( "scientific_name", str ),
                                   _data( "scientific_names", str ),  # array?
                                   _data( "synonym", str ),
                                   _data( "teleomorph", str ),
                                   _data( "type_material", str )) ) )

Sequence = _add( _node( label = "Sequence",
                        is_complete = False,
                        properties = [_data( "description", str, is_default = True )] ) )

Subsequence = _add( _node( "Subsequence" ) )

Go = _add( _node( "Go" ) )

Ipr = _add( _node( "Ipr" ) )

Kegg = _add( _node( "Kegg" ) )

Pfam = _add( _node( "Pfam" ) )

BlastQueriesSequence = _add( _edge( Blast, "Queries", Sequence ) )

BlastUsesSubsequence = _add( _edge( Blast, "Queries", Subsequence ) )

FamilyAccompaniesFamily = _add( _edge( Family, "Accompanies", Family ) )

FamilyAvoidsFamily = _add( _edge( Family, "Avoids", Family ) )

FamilyContainsSequence = _add( _edge( Family, "Contains", Sequence ) )

FamilyReachesTaxon = _add( _edge( Family, "Reaches", Taxon ) )

FeatureOverlapsFeature = _add( _edge( Feature, "Overlaps", Feature ) )

SourceContainsSequence = _add( _edge( Source, "Contains", Sequence ) )

GoClassContainsGoClass = _add( _edge( Go, "Contains", Go ) )

GoClassContainsSequence = _add( _edge( Go, "Contains", Sequence ) )

IprClassContainsSequence = _add( _edge( Ipr, "Contains", Sequence ) )

KeggClassContainsSequence = _add( _edge( Kegg, "Contains", Sequence ) )

PfamClassContainsSequence = _add( _edge( Pfam, "Contains", Sequence ) )

PlasmidContainsSequence = _add( _edge( Plasmid, "Contains", Sequence ) )

SequenceBlastqueriedSequence = _add( _edge( Sequence, "Like", Sequence ) )

SequenceContainsFeature = _add( _edge( Sequence, "Contains", Feature ) )

SequenceContainsSubsequence = _add( _edge( Sequence, "Contains", Subsequence ) )

SequenceReferencesReference = _add( _edge( Sequence, "References", Reference ) )

SourceContainsFamily = _add( _edge( Source, "Contains", Family ) )

SourceTouchesFamily = _add( _edge( Source, "Touches", Family ) )

SubsequenceBlastqueriedSubsequence = _add( _edge( Subsequence, "Like", Subsequence ) )

SubsequenceOverlapsSubsequence = _add( _edge( Subsequence, "Overlaps", Subsequence ) )

TaxonContainsSequence = _add( _edge( Taxon, "Contains", Sequence ) )

TaxonContainsTaxon = _add( _edge( Taxon, "Contains", Taxon ) )



"""
Defines and instantiates the application.
"""

import intermake
import neocommand
from bio42 import __version__
from bio42.application import help


class Application( neocommand.Application ):
    INSTANCE: "Application" = None
    
    
    def __init__( self, *args, **kwargs ):
        super().__init__( *args, **kwargs )
        
        from bio42 import schema
        schema.register( self.core.schema )
    
    
    def on_create_controller( self, mode: str ) -> intermake.Controller:
        if intermake.EImRunMode.is_gui( mode ):
            import bio42_gui
            return bio42_gui.B42GuiController()
        
        return super().on_create_controller( mode )


Application.INSTANCE = Application( inherit = neocommand.Application.INSTANCE,
                                    name = "bio42",
                                    version = __version__ )
app = Application.INSTANCE
help.init( Application.INSTANCE )



from os import path
from PyQt5.QtWidgets import QInputDialog, QWidget, QMessageBox
from mhelper import file_helper, io_helper, safe_cast
from mhelper_qt import exceptToGui, exqtSlot

import os
import webbrowser
import intermake
import neocommand

from bio42_gui.forms.designer import frm_database_designer
from bio42_gui.forms.frm_b42_base import AbstractB42SubWindow


_SS_WAIT = "busy"
_SS_ON = "online"
_SS_OFF = "offline"
_SS_ERROR = "error"


class FrmDatabase( AbstractB42SubWindow ):
    """
    Friendly GUI for the `transfer` command.
    """
    
    
    @exceptToGui()
    def __init__( self, parent: QWidget ):
        """
        CONSTRUCTOR
        """
        super().__init__( parent )
        self.ui = frm_database_designer.Ui_Dialog( self )
        self.ui.GRP_ENDPOINT.setVisible( False )
        self.ui.GRP_DATABASE.setVisible( False )
        self.setWindowTitle( "Database" )
        self.endpoint_list = []
        self.known_status = None
        self.status: neocommand.Neo4jStatus = None
        self.shown = False
    
    
    def on_shown( self ):
        self.update_status( None )
    
    
    @property
    def s_endpoint( self ):
        return self.status.endpoint if self.status is not None else None
    
    
    @property
    def s_database( self ):
        return self.status.database if self.status is not None else None
    
    
    def update_endpoints_list( self ):
        self.ui.CMB_ENDPOINT.clear()
        self.endpoint_list.clear()
        
        for endpoint in neocommand.get_core().endpoint_manager:
            if isinstance( endpoint, neocommand.DatabaseEndpoint ):
                self.ui.CMB_ENDPOINT.addItem( endpoint.name )
                self.endpoint_list.append( endpoint )
        
        if not self.endpoint_list:
            QMessageBox.warning( self, "Oops", "You haven't defined any database endpoints. You won't be able to query the database until you do so." )
        
        if self.s_endpoint is not None:
            try:
                index = self.endpoint_list.index( self.s_endpoint )
            except ValueError:
                index = -1
            
            self.ui.CMB_ENDPOINT.setCurrentIndex( index )
    
    
    def update_database_list( self, selection: str = None ):
        self.ui.CMB_DATABASE.clear()
        
        if self.s_endpoint is None:
            return
        
        src_dir = path.join( self.s_endpoint.directory, "data/databases" )
        items = []
        
        for dir in file_helper.list_dir( src_dir ):
            if not path.isdir( dir ):
                continue
            
            if not path.isfile( path.join( dir, "neostore" ) ):
                if len( os.listdir( dir ) ) != 0:
                    continue
            
            name = file_helper.get_filename( dir )
            
            self.ui.CMB_DATABASE.addItem( name )
            items.append( name )
        
        selection = selection or self.s_database
        
        if selection:
            try:
                index = items.index( selection )
            except ValueError:
                index = -1
            
            self.ui.CMB_DATABASE.setCurrentIndex( index )
    
    
    def update_status( self, target ):
        #
        # Select the first database if no `target` is specified
        #
        if target is None:
            for endpoint in neocommand.get_core().endpoint_manager:
                if isinstance( endpoint, neocommand.DatabaseEndpoint ):
                    target = endpoint
                    break
        
        #
        # Reset our information
        #
        self.status = None
        
        #
        # Clear the text if there is no suitable target
        #
        if target is None:
            self.ui.TXT_SERVER.setText( "(none)" )
            self.ui.TXT_DATABASE.setText( "N/A" )
            self.ui.TXT_STATUS.setText( "N/A" )
            self.set_label( "error", "no database selected" )
            return
        
        #
        # If there is a suitable target, display "please wait"
        #
        self.ui.TXT_SERVER.setText( "(please wait)" )
        self.ui.TXT_DATABASE.setText( "(please wait)" )
        self.ui.TXT_STATUS.setText( "(please wait)" )
        self.set_label( _SS_WAIT, "(please wait)" )
        
        #
        # Send the command asking for information
        #
        intermake \
            .acquire( neocommand.test_connection,
                      window = self,
                      auto_close = True,
                      callback = self.__test_connection_callback ) \
            .run( target )
    
    
    def __test_connection_callback( self, result: intermake.Result ):
        if result.is_error:
            return
        
        if result.command in (neocommand.set_database, neocommand.control_server):
            assert self.s_endpoint, "Didn't expect endpoint to change whilst database was being controlled."
            self.update_status( self.s_endpoint )
            return
        
        self.handle_status( safe_cast( neocommand.Neo4jStatus, result.result ) )
    
    
    def handle_status( self, status: neocommand.Neo4jStatus ):
        self.status = status
        self.ui.TXT_DATABASE.setText( status.database or "Unknown" )
        self.ui.TXT_SERVER.setText( status.endpoint.name )
        
        if status.error:
            self.set_label( _SS_ERROR, status.error )
        elif status.is_connected:
            self.set_label( _SS_ON, status.version )
        elif status.is_running:
            self.set_label( _SS_ERROR, "running but not connected (?) should have received an error code from neocommand but didn't." )
        else:
            self.set_label( _SS_OFF, status.version )
    
    
    def set_label( self, i, x ):
        self.known_status = i
        self.ui.TXT_STATUS.setText( i + ((" - " + x) if x else "") )
        self.ui.BTN_START.setEnabled( i is _SS_OFF )
        self.ui.BTN_STOP.setEnabled( i is _SS_ON )
        self.ui.BTN_REFRESH.setEnabled( i is not _SS_WAIT )
        self.ui.BTN_DATABASE.setEnabled( i is _SS_OFF )
        self.ui.BTN_ENDPOINT.setEnabled( i is not _SS_WAIT )
        self.ui.BTN_OPEN_BROWSER.setEnabled( i is _SS_ON )
        self.ui.BTN_SHOW_CONFIG.setEnabled( i is _SS_OFF or i is _SS_ERROR )
        
        if i is _SS_ON:
            self.ui.LBL_STATUS_COLOUR.setStyleSheet( "background: green;" )
        elif i is _SS_OFF:
            self.ui.LBL_STATUS_COLOUR.setStyleSheet( "background: darkred;" )
        else:
            self.ui.LBL_STATUS_COLOUR.setStyleSheet( "background: #C0C000;" )
    
    
    def set_display_page( self, i ):
        self.ui.GRP_ENDPOINT.setVisible( i == 1 and not self.ui.GRP_ENDPOINT.isVisible() )
        self.ui.GRP_DATABASE.setVisible( i == 2 and not self.ui.GRP_DATABASE.isVisible() )
    
    
    @exqtSlot()
    def on_BTN_START_clicked( self ) -> None:
        """
        Signal handler:
        """
        self.set_display_page( 0 )
        
        if self.known_status is not _SS_OFF:
            return
        
        if self.s_endpoint is None:
            return
        
        neocommand.control_server( self.s_endpoint, neocommand.ENeo4jCommand.START )
    
    
    @exqtSlot()
    def on_BTN_STOP_clicked( self ) -> None:
        """
        Signal handler:
        """
        self.set_display_page( 0 )
        
        if self.known_status is not _SS_ON:
            return
        
        if self.s_endpoint is None:
            return
        
        neocommand.control_server( self.s_endpoint, neocommand.ENeo4jCommand.STOP )
    
    
    @exqtSlot()
    def on_BTN_APPLYDB_clicked( self ) -> None:
        """
        Signal handler:
        """
        if self.known_status is not _SS_OFF:
            return
        
        if self.s_endpoint is None:
            return
        
        neocommand.set_database( self.s_endpoint, self.ui.CMB_DATABASE.currentText() )
    
    
    @exqtSlot()
    def on_BTN_OPEN_BROWSER_clicked( self ) -> None:
        """
        Signal handler:
        """
        if self.s_endpoint is None:
            return
        
        url = "http://{}:{}".format( self.s_endpoint.remote_address, self.s_endpoint.port )
        webbrowser.open( url )
    
    
    @exqtSlot()
    def on_BTN_SHOW_CONFIG_clicked( self ) -> None:
        """
        Signal handler:
        """
        if self.s_endpoint is None:
            return
        
        url = path.join( self.s_endpoint.directory, "conf", "neo4j.conf" )
        io_helper.system_select( url )
    
    
    @exqtSlot()
    def on_BTN_NEWDB_clicked( self ) -> None:
        """
        Signal handler:
        """
        if self.s_endpoint is None:
            return
        
        name, ok = QInputDialog.getText( self, "New database", "Please name your database" )
        
        if not name or not ok:
            return
        
        dir = path.join( self.s_endpoint.directory, "data", "databases", name )
        file_helper.create_directory( dir )
        
        self.update_database_list( name )
    
    
    @exqtSlot()
    def on_BTN_UPDATE_ENDPOINT_clicked( self ) -> None:
        """
        Signal handler:
        """
        self.set_display_page( 0 )
        
        i = self.ui.CMB_ENDPOINT.currentIndex()
        
        if i == -1:
            return None
        
        self.update_status( self.endpoint_list[i] )
    
    
    @exqtSlot()
    def on_BTN_ENDPOINT_clicked( self ) -> None:
        """
        Signal handler:
        """
        self.set_display_page( 1 )
        
        if self.ui.GRP_ENDPOINT.isVisible():
            self.update_endpoints_list()
    
    
    @exqtSlot()
    def on_BTN_DATABASE_clicked( self ) -> None:
        """
        Signal handler:
        """
        self.set_display_page( 2 )
        
        if self.ui.GRP_DATABASE.isVisible():
            self.update_database_list()
    
    
    @exqtSlot()
    def on_BTN_REFRESH_clicked( self ) -> None:
        """
        Signal handler:
        """
        self.set_display_page( 0 )
        
        if self.known_status is _SS_WAIT:
            return
        
        self.update_status( self.s_endpoint )



"""
Functions for applying annotations from the Uniprot online service.
"""
from typing import Optional
from mhelper import BatchList, SwitchError, MFlags
import neocommand
import intermake

from bio42.application import app
from bio42 import schema


@app.command( folder = "import" )
def download_accessions( endpoint: neocommand.Destination,
                         origin: neocommand.Origin,
                         input_type: str,
                         output_type: str = "ACC",
                         input_property: neocommand.isNodeProperty[schema.Sequence] = "uid",
                         input_column: Optional[int] = None,
                         output_property: neocommand.isNodeProperty[schema.Sequence] = "uniprot",
                         batch_size: int = 100 ) -> neocommand.Destination:
    """
    Uses UniProt to adds or replace accessions on gene Sequences.
    
    :param origin:              From where to retrieve the Sequences to update
    :param endpoint:            Where to send the new accessions to
    :param input_type:          Type of the input accession.
                                See http://www.uniprot.org/help/programmatic_access#id_mapping_examples
    :param output_type:         Type of the output accession.
    :param input_property:      Property to retrieve the input accession from.
                                For instance, `uid` or `accession`. 
    :param input_column:        Column of the input accession.
                                This only applies to pipe-delimited `input_property`s of the form `a|b|c`. 
    :param output_property:     Property to store the output accession into.
                                For instance, `uid` or `accession`.
    :param batch_size:          Number of requests to make to Uniprot at once. 
    :return:                    `endpoint` 
    """
    from urllib.parse import urlencode
    from urllib.request import Request
    from urllib.request import urlopen
    
    URL = "http://www.uniprot.org/uploadlists/"
    
    with origin.open_reader() as reader:
        to_do = reader.read_all_props( schema.Sequence.label, input_property )
        
    to_do = BatchList( to_do, batch_size )
    
    with endpoint.open_writer() as ep:
        with intermake.pr.pr_action( "Processing sequences", len( to_do ) ) as action:
            for next_portion in to_do:
                action.increment( len( next_portion ) )
                
                next_dict = { }
                
                for sequence_uid, property_value in next_portion:
                    if input_column is not None:
                        property_value = property_value.split( "|" )[input_column]
                    
                    next_dict[property_value] = sequence_uid
                
                data = urlencode( {
                    "from"  : input_type,
                    "to"    : output_type,
                    "format": "tab",
                    "query" : " ".join( next_dict.keys() )
                } )
                
                data = data.encode( 'ascii' )
                
                request = Request( URL, data )
                
                response = urlopen( request )
                tsv = response.read( 200000 ).decode( "utf-8" )
                
                lines = tsv.split( "\n" )[1:]  # first line just says "From-To"
                
                for line in lines:
                    accession_from, accession_to = line.split( "\t", 1 )
                    sequence_uid = next_dict[accession_from]  # type: str
                    
                    ep.create_node( label = schema.Sequence.label,
                                    uid = sequence_uid,
                                    properties = { output_property: accession_to } )
    
    return endpoint


class EAutoAnnotateMode( MFlags ):
    KEGG = 1
    GO = 2
    PFAM = 4


@app.command( folder = "import" )
def download_annotations( endpoint: neocommand.Destination,
                          origin: neocommand.Origin,
                          accession_property: str = "uniprot",
                          batch_size: int = 100,
                          create_edges: bool = True,
                          create_properties: bool = False,
                          create_nodes: bool = True,
                          property_prefix: str = "uniprot_",
                          modes: Optional[EAutoAnnotateMode] = None ):
    """
    Annotates sequences using Uniprot.
    Supports: PFAM, GO, KEGG.
    
    :param modes:                   What to apply. If not specified defaults to everything.
    :param create_properties:       Create properties for metadata. 
    :param create_edges:            Create edges for metadata.
    :param create_nodes:            Create nodes for metadata (requires `create_edges`).
    :param endpoint:                Where to send the results to
    :param origin:                  Where to get the sequences from
    :param accession_property:      Property to get the Uniprot accession from (on the `origin` sequences)
    :param batch_size:              Number of sequences to process at once. Note that Uniprot may also impose its own batching.
    :param property_prefix:         Prefix to add to the resultant properties (added to the `origin` sequences)
    :return:                        The `endpoint` argument.
    """
    import uniprot
    
    SEQUENCE_LABEL = schema.Sequence.label
    
    with origin.open_reader() as reader:
        to_do = reader.read_all_props( SEQUENCE_LABEL, accession_property )
    
    to_do = BatchList( to_do, batch_size )
    
    interesting = []
    
    if modes is None:
        modes = EAutoAnnotateMode.PFAM | EAutoAnnotateMode.GO | EAutoAnnotateMode.KEGG
    
    # noinspection PyTypeChecker
    for mode in modes:
        if mode.GO:
            interesting.append( ("go", schema.Go.label) )
        elif mode.PFAM:
            interesting.append( ("pfam", schema.Pfam.label) )
        elif mode.KEGG:
            interesting.append( ("kegg", schema.Kegg.label) )
        else:
            raise SwitchError( "modes[n]", mode )
    
    with endpoint.open_writer() as writer:
        with intermake.pr.pr_action( "Iterating sequences", len( to_do ) ) as action:
            for batch in to_do:
                action.increment( len( batch ) )
                accession_to_uid = dict( (v, k) for k, v in batch )
                
                sequence_id_string = " ".join( [x[1] for x in batch] )
                uniprot_data = uniprot.batch_uniprot_metadata( sequence_id_string, 'cache' )
                
                assert uniprot_data
                
                for sequence_accession, sequence_metadata in uniprot_data.items():
                    sequence_uid = accession_to_uid[sequence_accession]
                    
                    new_dict = { }
                    
                    for attribute, class_ in interesting:
                        if attribute in sequence_metadata:
                            value = sequence_metadata[attribute]
                            value = [x.split( ";", 1 )[0].strip() for x in value]
                            new_dict[property_prefix + attribute] = value
                            
                            if create_edges:
                                for mode in value:
                                    x_uid = mode
                                    
                                    if create_nodes:
                                        writer.create_node( label = class_, uid = x_uid, properties = { } )
                                    
                                    writer.create_edge( label = schema.GoClassContainsSequence,
                                                          start_uid = x_uid,
                                                          end_uid = sequence_uid,
                                                          properties = { } )
                    
                    if create_properties:
                        writer.create_node( label = SEQUENCE_LABEL, uid = sequence_uid, properties = new_dict )
    
    return endpoint



from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtGui import QShowEvent
from PyQt5.QtWidgets import QDialog, QWidget, QMessageBox
from mhelper import NOT_PROVIDED

import intermake


class AbstractB42SubWindow( QDialog ):
    __sig_shown = pyqtSignal()
    
    
    def __init__( self, parent: QWidget ):
        """
        CONSTRUCTOR
        """
        from bio42_gui.forms.frm_b42main import FrmB42Main
        self.main_window: FrmB42Main = parent
        super().__init__( parent )
        
        self.__sig_shown_emitted = False
        self.__sig_shown.connect( self.on_shown, Qt.QueuedConnection )
    
    
    def show_command( self, command, **kwargs ) -> intermake.Result:
        return self.main_window.controller.acquire( command, window = self, confirm = True ).run( **kwargs )
    
    
    def information( self, message ):
        QMessageBox.information( self,
                                 self.windowTitle(),
                                 message )
    
    
    def on_plugin_completed( self, result: intermake.Result ):
        pass
    
    
    def on_shown( self ):
        """
        This is called to allow the derived class to process any response after
        the dialogue is first shown and is visible to the user.
        """
        pass
    
    
    def on_get_result( self ):
        """
        This is called to allow the derived class to return any response to
        `request` having been called. If this is not implemented the result
        of `exec_` is returned.
        """
        return NOT_PROVIDED
    
    
    @classmethod
    def request( cls, parent: QWidget, *args, **kwargs ):
        from bio42_gui.forms.frm_b42main import FrmB42Main
        parent: FrmB42Main = parent
        # noinspection PyArgumentList
        frm = cls( parent, *args, **kwargs )
        parent.child_windows.append( frm )
        r = frm.exec_()
        parent.child_windows.remove( frm )
        x = frm.on_get_result()
        if x is not NOT_PROVIDED:
            r = x
        return r
    
    
    def showEvent( self, e: QShowEvent ):
        super().showEvent( e )
        
        if self.__sig_shown_emitted:
            return
        
        self.__sig_shown_emitted = True
        self.__sig_shown.emit()  # callback to be received AFTER the window is fully displayed
